/** \file alias.c
  Functions for storing and retrieving alias information.
*/
#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>


#include "config.h"
#include "util.h"
#include "alias.h"
#include "proc.h"
#include "parser.h"
#include "common.h"

/**
   Table containing all aliases
*/
static hash_table alias;

/**
   Struct describing an alias
*/
typedef struct
{
	/** Alias definition */
	wchar_t *cmd;
	/** Does the alias use variables */
	int use_vars;
}
alias_data;

/**
   Free all contents of an entry to the alias hash table
*/
static void clear_alias_entry( const void *key, 
							   const void *data )
{
	alias_data *d = (alias_data *)data;
	free( (void *)key);
	free( (void *)d->cmd );
	free( (void *)d );
}

void alias_init()
{
	hash_init( &alias,
			   &hash_wcs_func,
			   &hash_wcs_cmp );
}

void alias_destroy()
{
	hash_foreach( &alias, &clear_alias_entry );
	hash_destroy( &alias );
}


void alias_add( const wchar_t *name, 
				int use_vars,
				const wchar_t *val )
{
	if( alias_exists( name ) )
		alias_remove( name );
	
	alias_data *d = malloc( sizeof( alias_data ) );
	d->cmd = wcsdup( val );
	d->use_vars = use_vars;	
	hash_put( &alias, wcsdup(name), d );
}

int alias_exists( const wchar_t *cmd )
{
	return (hash_get(&alias, cmd) != 0 );
}

void alias_remove( const wchar_t *name )
{
	void *key;
	alias_data *d;
	if( !hash_contains( &alias,
						name ) )
		return;
	hash_remove( &alias,
				 name,
				 (const void **) &key,
				 (const void **)&d );
	
	free( key );
	free( d->cmd );
	free( d );
}

int alias_use_vars( const wchar_t *argv )
{
	alias_data *data = 
		(alias_data *)hash_get( &alias, argv );
	if( data == 0 )
		return -1;
	return data->use_vars;
}

	
const wchar_t *alias_get_definition( const wchar_t *argv )
{
	alias_data *data = 
		(alias_data *)hash_get( &alias, argv );
	if( data == 0 )
		return 0;
	return data->cmd;
}

	
void alias_get_names( array_list *list )
{
	hash_get_keys( &alias, list );
}
