/** \file builtin.c
	Functions for executing builtin functions.
*/
#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>


#include "config.h"
#include "util.h"
#include "wutil.h"
#include "builtin.h"
#include "alias.h"
#include "complete.h"
#include "proc.h"
#include "parser.h"
#include "reader.h"
#include "env.h"
#include "expand.h"
#include "common.h"
#include "wgetopt.h"
#include "sanity.h"
#include "tokenizer.h"
#include "builtin_help.h"
#include "wildcard.h"

/**
   Table of all builtins
*/
static hash_table builtin;

/**
   Wrapper for wcsfilecmp providing the correct signature for qsort
*/
static int mycmp( const void *a, const void *b )
{
	return wcsfilecmp( *(wchar_t **)a, *(wchar_t **)b );
}

/**
   Counts the number of non null pointers in the specified array
*/
static int count_args( wchar_t **argv )
{
	int argc = 1;
	while( argv[argc] != 0 )
	{
		argc++;
	}
	return argc;
}

/** 
	Buffers for storing the output of builtin functions
*/
string_buffer *sb_out=0, *sb_err=0;
/**
   Stack containing builting I/O for recursive builtin calls.
*/
static array_list io_stack;

/** 
	The file from which builtin functions should attempt to read, use instead of stdin.
*/
static FILE *builtin_stdin;

/** 
	This function works like wperror, but it prints its resuilt into the sb_err string_buffer instead of to stderr. Used by the builtin commands.
*/
static void builtin_wperror( const wchar_t *s)
{
	if( s != 0 )
	{
		sb_append2( sb_err, s, L": ", 0 );
	}
	char *err = strerror( errno );
	wchar_t *werr = str2wcs( err );
	sb_append2( sb_err,  werr, L"\n", 0 );	
	free( werr );
}


/*
  Here follows the definition of all builtin commands. The function
  names are all on the form builtin_NAME where NAME is the name of the
  builtin. so the function name for the builtin 'jobs' is
  'builtin_jobs'.

  Two builtins, 'command' and 'builtin' are not defined here as they
  are part of the parser. (They are not parsed as commands, instead
  they only slightly alter the parser state)

*/


/**
   Noop function. A fake function which sucessfully does nothing, for builtins which are handled by the parser, such as command and while. 
*/
static int builtin_ignore( wchar_t **argv )
{
	return 0;
}

/**
   Print help for the specified builtin
*/
static void builtin_print_help( wchar_t *cmd )
{
	const char *h = builtin_help_get( cmd );
	if( !h )
		return;
	
	wchar_t *str = str2wcs(builtin_help_get( cmd ));
	if( str )
	{
		sb_append( sb_err, str );
		free( str );
	}
}

/**
   The alias builtin, used for providing subroutines.
   It calls various functions from alias.c to perform any heavy lifting.
*/
static int builtin_alias( wchar_t **argv )
{	
	int i;
	int use_vars = 0, erase=0;

	array_list names;
	wchar_t **names_arr;

	int argc=count_args( argv );

	woptind=0;

	static struct woption
		long_options[] =
		{
			{
				L"variables", no_argument, 0, 'v'
			}
			,
			{
				L"erase", no_argument, 0, 'e' 
			}
			,
			{ 
				0, 0, 0, 0 
			}
		}
	;		
		
	while( 1 )
	{
		int opt_index = 0;
		
		int opt = wgetopt_long( argc,
								argv, 
								L"ve", 
								long_options, 
								&opt_index );
		if( opt == -1 )
			break;
			
		switch( opt )
		{
			case 0:
				if(long_options[opt_index].flag != 0)
					break;
				sb_append2( sb_err,
							L"alias: Unknown option ",
							long_options[opt_index].name,
							L"\n",
							0);				
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;
				
			case 'v':		
				use_vars=1;				
				break;

			case 'e':		
				erase=1;				
				break;

			case '?':
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;
				
		}
		
	}		

	if( erase )
	{
		int i;
		for( i=woptind; i<argc; i++ )
			alias_remove( argv[i] );
		return 0;
	}

	switch( argc - woptind )
	{
		case 0:
			al_init( &names );
			alias_get_names( &names );
			names_arr = list_to_char_arr( &names );
			qsort( names_arr, 
				   al_get_count( &names ), 
				   sizeof(wchar_t *), 
				   &mycmp );
			for( i=0; i<al_get_count( &names ); i++ )
			{
				sb_append2( sb_out,
							names_arr[i],
							(alias_use_vars(names_arr[i])?L" -v ":L" "),
							L"\"",
							alias_get_definition(names_arr[i]),
							L"\"\n",
							0);
			}
			free( names_arr );
			al_destroy( &names );			
			break;
		case 2:
			alias_add( argv[woptind], 
					   use_vars, 
					   argv[woptind+1]);
			break;
		default:
			sb_append( sb_err,
					   L"Syntax: alias [[-v] NAME VALUE]\n" );
			sb_append2( sb_err, 
						parser_current_line(),
						L"\n",
						0 );
				
			return 1;
			break;
	}
	return 0;

}

/**
   The read builtin. Reads from stdin and stores the values in environment variables.
*/
static int builtin_read( wchar_t **argv )
{
	wchar_t *buff=0;
	int buff_len = 0;
	int i, argc = count_args( argv );
	wchar_t *ifs;
	static int place = 0;
	wchar_t *nxt;
	
	woptind=0;

	while( 1 )
	{
		static struct woption
			long_options[] =
			{
				{
					L"export", no_argument, &place, ENV_EXPORT 
				}
				,
				{
					L"global", no_argument, &place, ENV_GLOBAL
				}
				,
				{ 
					0, 0, 0, 0 
				}
			}
		;		
		
		int opt_index = 0;
		
		int opt = wgetopt_long( argc,
								argv, 
								L"xg", 
								long_options, 
								&opt_index );
		if( opt == -1 )
			break;
			
		switch( opt )
		{
			case 0:
				if(long_options[opt_index].flag != 0)
					break;
				sb_append2( sb_err, 
							L"read: Unknown option ",
							long_options[opt_index].name,
							L"\n",
							0 );
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;
				
			case L'x':		
				place=ENV_EXPORT;
				break;
			case L'g':		
				place=ENV_GLOBAL;
				break;
			case L'?':
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;	
		}
		
	}		

	ifs = env_get( L"IFS" );
	if( ifs == 0 )
		ifs = L"";

	int len = fgetws2( &buff, &buff_len, builtin_stdin );
	if ( len == -1 )
	{
		if( buff )
			free( buff );
		builtin_wperror( L"read" );
		return 1;
	}
	
	i=woptind;
	wchar_t *state;
	
	nxt = wcstok( buff, (i<argc-1)?ifs:L"", &state );
	
	while( i<argc )
	{
		env_set( argv[i], nxt != 0 ? nxt: L"", place );		

		i++;
		if( nxt != 0 )
			nxt = wcstok( 0, (i<argc-1)?ifs:L"", &state);		
	}
	
	free( buff );
	return 0;
}

/**
   The eval builtin. Concatenates the arguments and calls eval on the result.
*/
static int builtin_eval( wchar_t **argv )
{
	wchar_t *tot, **ptr, *next;
	int totlen=0;

	for( ptr = argv+1; *ptr; ptr++ )
	{
		totlen += wcslen( *ptr) + 1;
	}
	tot = malloc( sizeof(wchar_t)*totlen );
	if( !tot )
	{
		sb_append( sb_err,
				   L"fish: out of memory\n" );
		sb_append2( sb_err, 
					parser_current_line(),
					L"\n",
					0 );		
		return(1);
	}
	for( ptr = argv+1, next=tot; *ptr; ptr++ )
	{
		int len = wcslen( *ptr );
		wcscpy( next, *ptr );
		next+=len;
		*next++=L' ';
	}
	*(next-1)=L'\0';
	eval( tot );
	free( tot );
	return proc_get_last_status();
}

/**
   The exit builtin. Calls do_exit to exit and returns the value specified.
*/
static int builtin_exit( wchar_t **argv )
{	
	int ec=0;
	do_exit();
	if( argv[1] != 0 )
		ec = wcstol(argv[1],0,10);
	return ec;
}

/**
   The cd builtin. Changes the current directory to the one specified
   or to $HOME if none is specified. The directory can be relative to
   any directory in the CDPATH variable.
*/
static int builtin_cd( wchar_t **argv )
{
	wchar_t *dir_in;
	wchar_t *dir;
	int free_dir = 0;


	if( argv[1]  == 0 )
		dir_in = env_get( L"HOME" );
	else 
		dir_in = argv[1];

	dir = parser_cdpath_get( dir_in );	

	if( !dir )
	{
		sb_append2( sb_err,
					L"cd: ",
					dir,
					L" is not a directory\n",
					0 );
		sb_append2( sb_err, 
					parser_current_line(),
					L"\n",
					0 );			
		return 1;
	}		


	char *dir_str = wcs2str( dir );

	if( chdir( dir_str ) != 0 )
	{
		free( dir_str );
		sb_append2( sb_err,
					L"cd: ",
					dir,
					L" is not a directory\n",
					0 );
		sb_append2( sb_err, 
					parser_current_line(),
					L"\n",
					0 );
		
		if( free_dir )
			free( dir );
		
		return 1;
	}
	free( dir_str );
	{
		char buf[4096];
		char *res = getcwd( buf, 4096 );
		wchar_t *dir_path;
		if( !res )
		{
			builtin_wperror( L"getcws" );
			return 1;
		}
		dir_path = str2wcs( res );
		
		env_set( L"PWD", dir_path, ENV_EXPORT );
		free( dir_path );
	}
	
	free( dir );

	return 0;
}

/**
   The complete builtin. Used for specifying programmable
   tab-completions. Calls the functions in complete.c for any heavy
   lifting.
*/
static int builtin_complete( wchar_t **argv )
{
	
	int argc=0;
	int result_mode=SHARED, long_mode=0;
	int cmd_type=-1;
	int remove = 0;
	int authorative = 1;
	
	wchar_t *cmd=0, short_opt=L'\0', *long_opt=L"", *comp=L"", *desc=L"";
	
	argc = count_args( argv );	
	
	woptind=0;
	
	while( 1 )
	{
		static struct woption
			long_options[] =
			{
				{
					L"exclusive", no_argument, 0, 'x' 
				}
				,
				{
					L"no-files", no_argument, 0, 'f' 
				}
				,
				{
					L"require-parameter", no_argument, 0, 'r' 
				}
				,
				{
					L"path", required_argument, 0, 'p'
				}
				,					
				{
					L"command", required_argument, 0, 'c' 
				}
				,					
				{
					L"short-option", required_argument, 0, 's' 
				}
				,
				{
					L"long-option", required_argument, 0, 'l'				}
				,
				{
					L"old-option", required_argument, 0, 'o' 
				}
				,
				{
					L"description", required_argument, 0, 'd'
				}
				,
				{
					L"arguments", required_argument, 0, 'a'
				}
				,
				{
					L"erase", no_argument, 0, 'e'
				}
				,
				{
					L"unauthorative", no_argument, 0, 'u'
				}
				,
				{ 
					0, 0, 0, 0 
				}
			}
		;		
		
		int opt_index = 0;
		
		int opt = wgetopt_long( argc,
								argv, 
								L"a:c:p:s:l:o:d:frxeu", 
								long_options, 
								&opt_index );
		if( opt == -1 )
			break;
			
		switch( opt )
		{
			case 0:
				if(long_options[opt_index].flag != 0)
					break;
				sb_append2( sb_err,
							L"complete: Unknown option ",
							long_options[opt_index].name,
							L"\n",
							0 );
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;
				
				
			case 'x':					
				result_mode |= EXCLUSIVE;
				break;
					
			case 'f':					
				result_mode |= NO_FILES;
				break;
				
			case 'r':					
				result_mode |= NO_COMMON;
				break;
					
			case 'p':					
				cmd_type = PATH;
				cmd = woptarg;
				break;
					
			case 'c':
				cmd_type = COMMAND;
				cmd = woptarg;
				break;
				
			case 'd':
				desc = woptarg;
				break;
				
			case 'u':
				authorative=0;
				break;
				
			case 's':
				if( wcslen( woptarg ) > 1 )
				{
					sb_append2( sb_err,
								L"complete: Parameter too long ",
								woptarg,
								L"\n",
								0);
					sb_append2( sb_err, 
								parser_current_line(),
								L"\n",
								0 );
					
					return 1;
				}
				
				short_opt = woptarg[0];
				break;
					
			case 'l':
				long_opt = woptarg;
				break;
				
			case 'o':
				long_mode=1;				
				long_opt = woptarg;
				break;

			case 'a':
				comp = woptarg;
				break;
				

			case 'e':
				remove = 1;
				
				break;
				

			case '?':
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;
				
		}
		
	}
	
	if( woptind != argc )
	{
		sb_append( sb_err, 
				   L"complete: Too many arguments\n" );
		sb_append2( sb_err, 
					parser_current_line(),
					L"\n",
					0 );
		return 1;
	}

	if( cmd == 0 )
	{
		/* No arguments specified, meaning we print the definitions of
		 * all specified completions to stdout.*/
		complete_print( sb_out );		
	}
	else
	{
		if( remove )
		{
			/* Remove the specified completion */
			complete_remove( cmd, 
							 cmd_type, 
							 short_opt,
							 long_opt );
		}
		else
		{
			/* Add the specified completion */
			complete_add( cmd, 
						  cmd_type, 
						  short_opt,
						  long_opt,
						  long_mode, 
						  result_mode, 
						  authorative,
						  comp,
						  desc ); 
		}
	}	
	return 0;
}

/**
   The source builtin. Can be called through either 'source' or
   '.'. Evaluates the contents of a file. 
*/
static int builtin_source( wchar_t ** argv )
{
	int stdin_org;
	int res;
	
	if( (argv[1] == 0) || (argv[2]!=0) )
	{
		sb_append( sb_err, L"Usage: . FILENAME\n" );
		sb_append2( sb_err, 
					parser_current_line(),
					L"\n",
					0 );
				
		return 1;
	}
	
	if( (stdin_org=dup( 0 )) == -1)
	{
		builtin_wperror(L"dup");
		return 1;
	}

	if( close( 0 ) )
	{
		builtin_wperror(L"close");
		return 1;
	}
	
/*	fwprintf( stderr, L"file %ls\n", argv[1] );*/
	
	if( wopen( argv[1], O_RDONLY ) == -1 )
	{		
		builtin_wperror( L"open" );
		res = 1;
	}
	else
	{
		reader_push_current_filename( argv[1] );
		res = reader_read();		
		if( close( 0 ) )
		{
			builtin_wperror(L"close");
			res = errno;
		}
		reader_pop_current_filename();
	}

	if( dup( stdin_org ) == -1)
	{
		builtin_wperror(L"dup");
		res = errno;
		fwprintf( stderr, L"Could not restore stdout\n" );
		sanity_lose();
	}
	
	if( close( stdin_org ) )
	{
		builtin_wperror(L"close");
		res = errno;
		fwprintf( stderr, L"Could not restore stdout\n" );
		sanity_lose();
	}

	return res;
}

/**
   The set builtin. Changes environment variables. Calls the functions in env.c for any heavy lifting.
*/
static int builtin_set( wchar_t **argv )
{
	int export=0, erase=0, global=0, local=0;
	int argc=count_args(argv);
	
	woptind=0;
	
	while( 1 )
	{
		struct woption
			long_options[] =
			{
				{
					L"export", no_argument, &export, 1 
				}
				,
				{
					L"global", no_argument, &global, 1 
				}
				,
				{
					L"local", no_argument, &local, 1 
				}
				,
				{
					L"erase", no_argument, &erase, 1 
				}
				,
				{ 
					0, 0, 0, 0 
				}
			}
		;		
		
		int opt_index = 0;
		
		int opt = wgetopt_long( argc,
								argv, 
								L"xgel", 
								long_options, 
								&opt_index );
		if( opt == -1 )
			break;
			
		switch( opt )
		{
			case 0:
				if(long_options[opt_index].flag != 0)
					break;
				
				sb_append2( sb_err, L"fish: Unknown option ",
							long_options[opt_index].name,
							L"\n",
							0 );				
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;
				
			case 'e':		
				erase = 1;
				break;
			case 'x':		
				export = 1;
				break;
			case 'l':		
				local = 1;
				break;
			case 'g':		
				global = 1;
				break;
			case '?':
				sb_append2( sb_err, 
							parser_current_line(),
							L"\n",
							0 );
				
				return 1;	
		}		
	}	

	if( local && global )
	{
			sb_append( sb_err, L"set: Variable can\'t be both global and local.\n" );
			builtin_print_help( argv[0] );
			return 1;
	}	

	if( erase )
	{
		if( woptind != argc-1 )
		{
			sb_append( sb_err, L"set: You must specify the name of the variable to erase\n" );
			sb_append2( sb_err, 
						parser_current_line(),
						L"\n",
						0 );
			builtin_print_help( argv[0] );
				
			return 1;
		}
		env_remove( argv[woptind], ENV_USER );
	}
	else
	{
		switch( argc-woptind )
		{
			case 0:
			{
				/* 
				   Print all enviroment variables
				*/
				array_list names;
				wchar_t **names_arr;
				int i;
				
				al_init( &names );
				env_get_names( &names );
				names_arr = list_to_char_arr( &names );
				qsort( names_arr, 
					   al_get_count( &names ), 
					   sizeof(wchar_t *), 
					   &mycmp );
				for( i=0; i<al_get_count( &names ); i++ )
				{
					wchar_t *key = expand_escape( names_arr[i], 1);
					wchar_t *val = expand_escape( env_get( names_arr[i] ), 1);
					
					sb_append2( sb_out,
								key,
								L" ",
								val,
								L"\n",
								0 );
					free( key );
					free(val);
				}
				free( names_arr );
				al_destroy( &names );			
				break;
			}
			case 1:
				env_set( argv[woptind], 0, (export?ENV_EXPORT:0)|(global?ENV_GLOBAL:0) );
				break;
				
			case 2:
				env_set( argv[woptind], argv[woptind+1], (local?ENV_LOCAL:0)|(export?ENV_EXPORT:0)|(global?ENV_GLOBAL:0)|ENV_USER);
				break;
				
			default:
				builtin_print_help( argv[0] );
				return 1;
		}
		
	}
	return 0;		
}

/**
   Make the specified job the first job of the job list. Moving jobs
   around in the list makes the list reflect the order in which the
   jobs where used.
*/
static void make_first( job_t *j )
{
	job_t *prev=0;
	job_t *curr;
	for( curr = first_job; curr != j; curr = curr->next )
	{
		prev=curr;
	}
	if( curr == j )
	{
		if( prev == 0 )
			return;
		else
		{
			prev->next = curr->next;
			curr->next = first_job;
			first_job = curr;
		}
	}
}


/**
   Builtin for putting a job in the foreground
*/
static int builtin_fg( wchar_t **argv )
{
	job_t *j;
	
	if( argv[1] == 0 )
	{
		for( j=first_job; ((j!=0) && (!j->constructed)); j=j->next )
			;
	}
	else if( argv[2] != 0 )
	{
		int pid = wcstol( argv[1], 0, 10 );
		j = job_get_from_pid( pid );
		if( j != 0 )
		{
			sb_append( sb_err, L"fg: Ambigous job\n" );	
		}
		else
		{
			sb_append2( sb_err, L"fg: Not a job (", argv[1], L")\n", 0 );
		}
		
		return 1;
	}
	else
	{
		int pid = wcstol( argv[1], 0, 10 );
		j = job_get_from_pid( pid );
	}

	if( j == 0 )
	{
		sb_append( sb_err, L"fg: No suitable job\n" );
		return 1;
	}
/*
  It would be nice to print some message before putting the job into
  the foreground, but we can't since the output won't be written until
  after the builtin returns, by which time the command will have
  exited. Maybe we should check if the user is redirecting output and if not print directly to screen.
*/
/*	else
	{
	sb_append( sb_err, L"Set job ");
	sb_append_int( sb_err, j->job_id );
	sb_append( sb_err, L" with pgid " );
	sb_append_int( sb_err, j->pgid );
	sb_append( sb_err, L" to fg\n" );
	}
*/

	wchar_t *ft = tok_first( j->command );
	if( ft != 0 )
		env_set( L"_", ft, ENV_EXPORT );
	free(ft);
	reader_write_title();
/*
	fwprintf( stderr, L"Send job %d, \'%ls\' to foreground\n", 
			  j->job_id,
			  j->command );
*/
	make_first( j );
	j->fg=1;
	
		
	job_continue( j, job_is_stopped(j) );
	return 0;
}

/**
   Helper function for builtin_bg()
*/
static void send_to_bg( job_t *j, wchar_t *name )
{
	if( j == 0 )
	{
		sb_append2( sb_err, L"bg: Unknown job ", name, "\n", 0 );
		return;
	}	
	else
	{
		sb_append( sb_err, L"Send job ");
		sb_append_int( sb_err, j->job_id );
		sb_append( sb_err, L", \'");
		sb_append( sb_err, j->command);
		sb_append( sb_err, L"\' to background\n" );
	}
	make_first( j );
	j->fg=0;
	job_continue( j, job_is_stopped(j) );
}


/**
   Builtin for putting a job in the background
*/
static int builtin_bg( wchar_t **argv )
{
	if( argv[1] == 0 )
	{
  		job_t *j;
		for( j=first_job; ((j!=0) && (!j->constructed) && (!job_is_stopped(j))); j=j->next )
			;
		send_to_bg( j, L"(default)");
		return 0;
	}
	for( argv++; *argv != 0; argv++ )
	{
		int pid = wcstol( *argv, 0, 10 );
		send_to_bg( job_get_from_pid( pid ), *argv);
	}
	return 0;
}


#ifdef HAVE__PROC_SELF_STAT
/**
   Calculates the cpu usage (in percent) of the specified job.
*/
static int cpu_use( job_t *j )
{
	double u=0;
	process_t *p;
	
	for( p=j->first_process; p; p=p->next )
	{
		struct timeval t;
		int jiffies;
		gettimeofday( &t, 0 );
		jiffies = proc_get_jiffies( p );
		
		double t1 = 1000000.0*p->last_time.tv_sec+p->last_time.tv_usec;
		double t2 = 1000000.0*t.tv_sec+t.tv_usec;
		
/*		fwprintf( stderr, L"t1 %f t2 %f p1 %d p2 %d\n",
				  t1, t2, jiffies, p->last_jiffies );
*/	

		u += ((double)(jiffies-p->last_jiffies))/(t2-t1);
	}
	return u*1000000;
}
#endif

/**
   Builtin for printing running jobs
*/
static int builtin_jobs( wchar_t **argv )
{	
	job_t *j;
	int found=0;
	
	for( j= first_job; j; j=j->next )
	{
		/*
		  Ignore unconstructed jobs, i.e. ourself.
		*/
		if( j->constructed )
		{
			if( !found )
			{
				/*
				  Print table header before first job
				*/
				sb_append( sb_out, L"Job\tGroup\t");
#ifdef HAVE__PROC_SELF_STAT
				sb_append( sb_out, L"CPU\t" );
#endif
				sb_append( sb_out, L"State\tCommand\n" );
			}
			
			found = 1;
			sb_append_int( sb_out, j->job_id );			
			sb_append( sb_out, L"\t" );
			sb_append_int( sb_out, j->pgid );
			sb_append( sb_out, L"\t" );
			
#ifdef HAVE__PROC_SELF_STAT
			sb_append_int( sb_out, cpu_use(j) );
			sb_append( sb_out, L"%\t" );
#endif
			sb_append2( sb_out, job_is_stopped(j)?L"stopped\t":L"running\t", 
						j->command, L"\n", 0 );
			
		}
	}
	if( !found )
	{
		sb_append( sb_out, L"jobs: There are no currently running jobs\n" );
	}
	return 0;
}

/**
   Builtin for looping over a list
*/
static int builtin_for( wchar_t **argv )
{
	int argc = count_args( argv );

	if( argc < 3) 
	{
		sb_append( sb_err, 
				   L"for: Expected at least two arguments\n" );				
		builtin_print_help( argv[0] );
		return 1;
	}
	else if (wcscmp( argv[2], L"in") != 0 )
	{
		sb_append( sb_err, 
				   L"for: Second argument must be \'in\'\n" );				
		builtin_print_help( argv[0] );
		return 1;		
	}
	else
	{
		parser_push_block( FOR );
		al_init( &current_block->for_vars);
		
		int i;
		current_block->tok_pos = parser_get_pos();
		current_block->for_variable = wcsdup( argv[1] );
		
		for( i=argc-1; i>3; i-- )
		{
			al_push( &current_block->for_vars, wcsdup(argv[ i ] ));
		}
		if( argc > 3 )
		{
			env_set( current_block->for_variable, argv[3], 0);
		}
		else
		{
			current_block->skip=1;
		}		
	}
	return 0;
}

/**
   Builtin for ending a block of code, such as a for-loop or an if statement
*/
static int builtin_end( wchar_t **argv )
{
	if( current_block == 0 )
	{
		sb_append( sb_err,
				   L"end: Not inside of block\n" );
		builtin_print_help( argv[0] );
		return 1;
	}
	else
	{
		int kill_block = 1;
		
		switch( current_block->type )
		{
			case WHILE:

				if( !( current_block->skip && (current_block->loop_status != LOOP_CONTINUE )))
				{
					current_block->loop_status = LOOP_NORMAL;
					current_block->skip = 0;
					kill_block = 0;
					parser_set_pos( current_block->tok_pos);
					current_block->while_state = WHILE_TEST_AGAIN;
				}
				
				break;
				
			case IF:
				break;

			case FOR:
				if( current_block->loop_status == LOOP_BREAK )
				{
					while( al_get_count( &current_block->for_vars ) )
					{
						free( (void *)al_pop( &current_block->for_vars ) );
					}
				}
				
				if( al_get_count( &current_block->for_vars ) )
				{
					wchar_t *val = (wchar_t *)al_pop( &current_block->for_vars );
					env_set( current_block->for_variable, val, 0);
					current_block->loop_status = LOOP_NORMAL;
					current_block->skip = 0;
					free(val);
					
					kill_block = 0;
					parser_set_pos( current_block->tok_pos );
/*					
				fwprintf( stderr, 
				L"jump to %d\n",
				current_block->tok_pos );								*/
				}
				break;
		}
		if( kill_block )
		{
			parser_pop_block();
		}
		return 0;
	}	
}

/**
   Builting for executing commands if an if statement is false
*/
static int builtin_else( wchar_t **argv )
{
	if( current_block == 0 || 
		current_block->type != IF ||
		current_block->if_state != 1)
	{
		sb_append( sb_err,
				   L"else: not inside of if block\n" );
		builtin_print_help( argv[0] );
		return 1;
	}
	else
	{
		current_block->if_state++;
		current_block->skip = !current_block->skip;
		env_pop();
		env_push(0);
	}
	return 0;
}

/**
   this function handles both the 'continue' and the 'break' builtins.
*/
static int builtin_break_continue( wchar_t **argv )
{
	int is_break = (wcscmp(argv[0],L"break")==0);
	int argc = count_args( argv );
	
	block_t *b = current_block;
	
	if( argc != 1 )
	{
		sb_append2( sb_err, 
					argv[0], 
					L": Unknown option \'", argv[1], L"\'", 0 );
		builtin_print_help( argv[0] );
		return 1;		
	}
	

	while( (b != 0) && 
		   ( b->type != WHILE) && 
		   (b->type != FOR ) )
	{
		b = b->outer;
	}
	if( b == 0 )
	{
		sb_append2( sb_err, 
					argv[0], 
					L": Not inside of loop\n", 0 );
		builtin_print_help( argv[0] );
		return 1;
	}
	
	b = current_block;
	while( ( b->type != WHILE) && 
		   (b->type != FOR ) )
	{
		b->skip=1;
		b = b->outer;
	}
	b->skip=1;
	b->loop_status = is_break?LOOP_BREAK:LOOP_CONTINUE;
	return 0;
}

/**
   Builting for executing one of several blocks of commands depending on the value of an argument.
*/
static int builtin_switch( wchar_t **argv )
{
	int argc = count_args( argv );
	
	if( argc != 2 )
	{
		sb_append( sb_err, L"switch: syntax error, expected exactly onw argument to switch\n" );
		builtin_print_help( L"switch" );
		return 1;
	}
	
	parser_push_block( SWITCH );
	current_block->switch_value = wcsdup( argv[1]);
	current_block->skip=1;
	current_block->switch_taken=0;

	return 0;
}

/**
   Builtin used together with the switch builtin
*/
static int builtin_case( wchar_t **argv )
{
	int argc = count_args( argv );
	int i;
	
	if( current_block->type != SWITCH )
	{
		sb_append( sb_err, L"case: syntax error, case command while not in switch block\n" );
		builtin_print_help( L"case" );
		return 1;
	}
		
	current_block->skip = 1;

	if( current_block->switch_taken )
	{
		return 0;
	}
	
	for( i=1; i<argc; i++ )
	{
		if( wildcard_match( current_block->switch_value, argv[i], 0 ) )
		{
			current_block->skip = 0;
			current_block->switch_taken = 1;
			break;		
		}
	}
	
	return 0;
		
}


/*
  END OF BUILTIN COMMANDS
  Below are functions for handling the builtin commands
*/


void builtin_init()
{
	al_init( &io_stack );
	hash_init( &builtin, &hash_wcs_func, &hash_wcs_cmp );
	hash_put( &builtin, L"exit", &builtin_exit );	
	hash_put( &builtin, L"cd", &builtin_cd );	
	hash_put( &builtin, L"alias", &builtin_alias );	
	hash_put( &builtin, L"complete", &builtin_complete );	
	hash_put( &builtin, L"end", &builtin_end );
	hash_put( &builtin, L"else", &builtin_else );
	hash_put( &builtin, L"eval", &builtin_eval );	
	hash_put( &builtin, L"for", &builtin_for );	
	hash_put( &builtin, L".", &builtin_source );	
	hash_put( &builtin, L"set", &builtin_set );	
	hash_put( &builtin, L"fg", &builtin_fg );	
	hash_put( &builtin, L"bg", &builtin_bg );	
	hash_put( &builtin, L"jobs", &builtin_jobs );	
	hash_put( &builtin, L"read", &builtin_read );	
	hash_put( &builtin, L"break", &builtin_break_continue );	
	hash_put( &builtin, L"continue", &builtin_break_continue );	

	hash_put( &builtin, L"switch", &builtin_switch );	
	hash_put( &builtin, L"case", &builtin_case );	
	
	/* 
	   Builtins that are handled directly by the parser. They are
	   bound to a noop function only so that help requests will be
	   honored.
	*/
	hash_put( &builtin, L"builtin", &builtin_ignore );	
	hash_put( &builtin, L"command", &builtin_ignore );		
	hash_put( &builtin, L"if", &builtin_ignore );	
	hash_put( &builtin, L"while", &builtin_ignore );	
	
	builtin_help_init();
}

void builtin_destroy()
{
	al_destroy( &io_stack );
	hash_destroy( &builtin );
	builtin_help_destroy();
}

int builtin_exists( wchar_t *cmd )
{
	return (hash_get(&builtin, cmd) != 0 );
}

void builtin_exec( wchar_t **argv, int type )
{
	switch( type )
	{
		case BUILTIN_REGULAR:
		{
			int (*cmd)(wchar_t **argv)=0;
			cmd = hash_get( &builtin, argv[0] );
			
			if( argv[1] != 0 )
			{

				if( argv[2] == 0 && ((wcscmp( argv[1], L"--help" )==0) || (wcscmp( argv[1], L"-h")==0)) )
				{
					builtin_print_help( argv[0] );
					return;
				}
			}
	
			if( cmd != 0 )
			{
				int count = 0;
				while( argv[count] != 0 )
					count++;	
		
				proc_set_last_status( cmd( argv ) );		
			}
			else
			{
				sb_append2( sb_err, L"fish: Unknown builtin: ", argv[0], 0 );	
			}
		}
		break;
		case BUILTIN_ALIAS:
		{
			wchar_t **arg;
			int i;
			const wchar_t * def = alias_get_definition( argv[0] );
			if( def == 0 )
			{
				fwprintf( stderr, L"Unknown alias %ls\n", argv[0] );
				sanity_lose();
				break;
			}
			env_push(1);
			for( i=1,arg = argv+1; *arg; i++, arg++ )
			{
				wchar_t var[16];
				swprintf( var, 16, L"%d", i );	
				env_set( var, *arg, 0 );
			}
			parser_forbid_alias( argv[0] );
			eval( def );
			parser_allow_alias();
			env_pop();
		}
		break;
		default:
			fwprintf( stderr, L"Called builtin_exec with invalid type %d\n", type );
			sanity_lose();
			break;
	}	
}

void builtin_get_names( array_list *list )
{
 	hash_get_keys( &builtin, list );
	al_push( list, L"command" );
	al_push( list, L"builtin" );	
}

void builtin_push_io( FILE *bi)
{
	if( builtin_stdin != 0 )
	{
		al_push( &io_stack, builtin_stdin );
		al_push( &io_stack, sb_out );
		al_push( &io_stack, sb_err );		
	}
	builtin_stdin = bi;
	sb_out = malloc(sizeof(string_buffer));
	sb_err = malloc(sizeof(string_buffer));
	sb_init( sb_out );
	sb_init( sb_err );	
}

void builtin_pop_io()
{
	builtin_stdin = 0;
	sb_destroy( sb_out );
	sb_destroy( sb_err );
	free( sb_out);
	free(sb_err);
	
	if( al_get_count( &io_stack ) >0 )
	{
		sb_err = (string_buffer *)al_pop( &io_stack );
		sb_out = (string_buffer *)al_pop( &io_stack );
		builtin_stdin = (FILE *)al_pop( &io_stack );
	}
	else
	{
		sb_out = sb_err = 0;
		builtin_stdin = 0;
	}
}

