/**\file expand.c

   String expantion functions. These functions perform several kinds of parameter expantion.
*/

#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <string.h>
#include <wctype.h>
#include <errno.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <termios.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>

#ifdef SunOS
#include <procfs.h>
#endif

#include "config.h"
#include "util.h"
#include "common.h"
#include "wutil.h"
#include "env.h"
#include "parser.h"
#include "expand.h"
#include "wildcard.h"
#include "proc.h"
#include "exec.h"
#include "tokenizer.h"
#include "complete.h"

/**
   Description for child process
*/
#define COMPLETE_CHILD_PROCESS_DESC COMPLETE_SEP_STR L"Child process"
/**
   Description for non-child process
*/
#define COMPLETE_PROCESS_DESC COMPLETE_SEP_STR L"Process"
/**
   Description for long job
*/
#define COMPLETE_JOB_DESC COMPLETE_SEP_STR L"Job"
/**
   Description for short job. The job command is concatenated
*/
#define COMPLETE_JOB_DESC_VAL COMPLETE_SEP_STR, L"Job: "

/**
 Return the environment variable value for the string starting at in 
*/
static wchar_t *expand_var( wchar_t *in )
{
	wchar_t *foo = in[0] == VARIABLE_EXPAND ? env_get( expand_var(in+1) ) : env_get( in );
	return ( foo == 0 )?L"":foo;
}

wchar_t *expand_escape( const wchar_t *in, int escape_all )
{
	wchar_t *out = malloc( sizeof(wchar_t)*(wcslen(in)*2 + 1));
	if( out != 0 )
	{
		wchar_t *pos=out;
		while( *in != 0 )
		{
			switch( *in )
			{
				case L'\t':
					*(pos++) = L'\\';
					*(pos++) = L't';					
					break;
					
				case L'\n':
					*(pos++) = L'\\';
					*(pos++) = L'n';					
					break;
					
				case L'\b':
					*(pos++) = L'\\';
					*(pos++) = L'b';					
					break;
					
				case L'\r':
					*(pos++) = L'\\';
					*(pos++) = L'r';					
					break;
					
				case L'\e':
					*(pos++) = L'\\';
					*(pos++) = L'e';					
					break;
					
				case L'\\':
				case L'&':
				case L'$':
				case L' ':
				case L'#':
				case L'^':
				case L'<':
				case L'>':
				case L'@':
				case L'(':
				case L')':
				case L'{':
				case L'}':
				case L'?':
				case L'*':
				case L'|':
				case L';':
				case L':':
					if( escape_all )
						*pos++ = L'\\';
				default:
					*pos++ = *in;
					break;
			}
			in++;
		}
		*pos = 0;
	}
	return out;
}

/**
   Tests if all characters in the string are numeric
*/
static int isnumeric( const char *n )
{
	if( *n == '\0' )
		return 1;
	if( *n < '0' || *n > '9' )
		return 0;
	return isnumeric( n+1 );
}

/**
   Tests if all characters in the wide string are numeric
*/
static int iswnumeric( const wchar_t *n )
{
	if( *n == L'\0' )
		return 1;
	if( *n < L'0' || *n > L'9' )
		return 0;
	return iswnumeric( n+1 );
}

/**
   See if the process described by \c proc matches the commandline \c cmd
*/
static int match_pid( const wchar_t *cmd, 
					  const wchar_t *proc,
					  int accept_incomplete )
{
	/* Test for direct match */
	
	if( wcsncmp( cmd, proc, wcslen( proc ) ) == 0 )
		return 1;
	
	if( accept_incomplete )
		return 0;

	/* 
	   Test if the commandline is a path to the command, if so we try
	   to match against only the command part
	*/
	wchar_t *first_token = tok_first( cmd );
	if( first_token == 0 )
		return 0;
		
	wchar_t *start=0;
	wchar_t prev=0;
	wchar_t *p;
	
	/* 
	   This should be done by basename(), if it wasn't for the fact
	   that is does not accept wide strings
	*/
	for( p=first_token; *p; p++ )
	{
		if( *p == L'/' && prev != L'\\' )
			start = p;
		prev = *p;
	}
	if( start )
	{
			
		if( wcsncmp( start+1, proc, wcslen( proc ) ) == 0 )
		{
			free( first_token );
			return 1;
		}
	}
	free( first_token );
	
	return 0;
	
}

/**
   Searches for a job with the specified job id, or a job or process
   which has the string \c proc as a prefix of its commandline.

   If accept_incomplete is true, the remainding string for any matches are inserted.

   If accept_incomplete is false, any job matching the specified
   string is matched, and the job pgid is returned. If no job
   matchess, all child processes are searched. If no child processes
   match, and <tt>fish</tt> can understand the contents of the /proc
   filesystem, all the users processes are searched for matches.
*/

static int find_process( const wchar_t *proc, 
						 int accept_incomplete, 
						 array_list *out )
{
	DIR *dir;
	struct dirent *next;
	char *pdir_name;
	char *pfile_name;
	wchar_t *cmd=0;
	int sz=0;
	int found = 0;
	wchar_t *result;	
	
	job_t *j;

	if( iswnumeric(proc ) || (wcslen(proc)==0) )
	{
		//	fwprintf( stderr, L"Numeric\n\n\n" );

		if( accept_incomplete )
		{
			for( j=first_job; j != 0; j=j->next )
			{
				wchar_t jid[16];
				if( j->command == 0 )
					continue;

				swprintf( jid, 16, L"%d", j->job_id );
//				fwprintf( stderr, L"Jid %ls\n", jid );
				
				if( wcsncmp( proc, jid, wcslen(proc ) )==0 )
				{
					al_push( out, wcsdupcat2( jid+wcslen(proc), COMPLETE_JOB_DESC_VAL, j->command, 0 ) );
					
					
				}
			}
			
		}
		else
		{
	
			int jid = wcstol( proc, 0, 10 );
		
			j = job_get( jid );
			if( (j != 0) && (j->command != 0 ) )
			{
				
				{
					result = malloc(sizeof(wchar_t)*16 );
					swprintf( result, 16, L"%d", j->pgid );
					//fwprintf( stderr, L"pid %d %ls\n", j->pgid, result );
					al_push( out, result );
					found = 1;
				}
			}		
		}
	}
	if( found )
		return 1;

	for( j=first_job; j != 0; j=j->next )
	{
//		fwprintf( stderr, L"..." );
		if( j->command == 0 )
			continue;
		
		//	fwprintf( stderr, L"match '%ls' '%ls'\n\n\n", j->command, proc );
		
		if( match_pid( j->command, proc, accept_incomplete ) )
		{
			if( accept_incomplete )
			{
				wchar_t *res = wcsdupcat( j->command + wcslen(proc), 
										  COMPLETE_JOB_DESC );
//				fwprintf( stderr, L"Woot %ls\n", res );
				
				al_push( out, res );
			}
			else
			{
				result = malloc(sizeof(wchar_t)*16 );
				swprintf( result, 16, L"%d", j->pgid );
				al_push( out, result );
				found = 1;
			}
		}
	}

	if( found )
	{
		return 1;
	}

	for( j=first_job; j; j=j->next )
	{
		process_t *p;
		if( j->command == 0 )
			continue;
		for( p=j->first_process; p; p=p->next )
		{
			
//		fwprintf( stderr, L"..." );
			if( p->actual_cmd == 0 )
				continue;
			
			//	fwprintf( stderr, L"match '%ls' '%ls'\n\n\n", j->command, proc );
			
			if( match_pid( p->actual_cmd, proc, accept_incomplete ) )
			{
				if( accept_incomplete )
				{
					wchar_t *res = wcsdupcat( p->actual_cmd + wcslen(proc), 
											  COMPLETE_CHILD_PROCESS_DESC );
					al_push( out, res );
				}
				else
				{
					result = malloc(sizeof(wchar_t)*16 );
					swprintf( result, 16, L"%d", p->pid );
					al_push( out, result );
					found = 1;
				}
			}
		}
	}

	if( found )
	{
		return 1;
	}

	if( !(dir = opendir( "/proc" )))
	{
		/*
		  This sytem does not have a /proc filesystem. 
		*/
		return 1;
	}
	
	pdir_name = malloc( 256 );
	pfile_name = malloc( 64 );
	strcpy( pdir_name, "/proc/" );
	
	while( (next=readdir(dir))!=0 )
	{
		char *name = next->d_name;
		struct stat buf;

		if( !isnumeric( name ) )
			continue;

		strcpy( pdir_name + 6, name );
		if( stat( pdir_name, &buf ) )
		{
			continue;
		}		
		if( buf.st_uid != getuid() )
		{
			continue;
		}
		strcpy( pfile_name, pdir_name );
		strcat( pfile_name, "/cmdline" );
		
		if( !stat( pfile_name, &buf ) )
		{
			/*
			  the 'cmdline' file exists, it should contain the commandline
			*/
			FILE *cmdfile;
			
			if((cmdfile=fopen( pfile_name, "r" ))==0)
			{
				wperror( L"fopen" );
				continue;
			}
			
			fgetws2( &cmd, &sz, cmdfile );		
			
			fclose( cmdfile );
		}
		else
		{
#ifdef SunOS
			strcpy( pfile_name, pdir_name );
			strcat( pfile_name, "/psinfo" );
			if( !stat( pfile_name, &buf ) )
			{
				psinfo_t info;

				FILE *psfile;
				
				if((psfile=fopen( pfile_name, "r" ))==0)
				{
					wperror( L"fopen" );
					continue;
				}
				
				if( fread( &info, sizeof(info), 1, psfile ) )
				{
					if( cmd != 0 )
						free( cmd );
					cmd = str2wcs( info.pr_fname );
				}
				fclose( psfile );
			}
			else
#endif
			{
				if( cmd != 0 )
				{
					*cmd = 0;
				}
			}
		}
		
		if( cmd != 0 )
		{
			if( match_pid( cmd, proc, accept_incomplete ) )
			{			
				if( accept_incomplete )
				{
					wchar_t *res = wcsdupcat( cmd + wcslen(proc),
											  COMPLETE_PROCESS_DESC );
					al_push( out, res );
				}
				else
				{
					al_push( out, str2wcs( name ) );
				}
			}
		}
	}

	if( cmd != 0 )
		free( cmd );

	free( pdir_name );
	free( pfile_name );

	closedir( dir );	

	return 1;
}

/**
   Process id expantion
*/
static int expand_pid( wchar_t *in, 
					   int accept_incomplete,
					   array_list *out )
{
	if( *in != PROCESS_EXPAND )
	{
		al_push( out, in );
		return 1;
	}
	
//	fwprintf( stderr, L"expand_pid() %ls\n", in );
	int prev = al_get_count( out );
	if( !find_process( in+1, accept_incomplete, out ) )
		return 0;
	
	if( prev == al_get_count( out ) )
	{
//		fwprintf( stderr, L"no match\n" );
		
		if( accept_incomplete )
			free( in );
		else
		{
			*in = L'%';
//			fwprintf( stderr, L"return %ls\n", in );
			al_push( out, in );
		}
	}	
	else
	{
//		fwprintf( stderr, L"match\n" );
		free( in );
	}
	
	return 1;
}


/**
  Expand all environment variables in the string *ptr. 
*/
static int expand_variables( wchar_t **ptr )
{
	wchar_t c;
	wchar_t prev_char=0;
	int i;
	wchar_t *in=*ptr;
	
	for( i=0; in[i]!= '\0'; i++ )
	{
		c = in[i];
		if( c == VARIABLE_EXPAND && prev_char != L'\\' )
		{
			int start_pos = i+1;
			int stop_pos = start_pos;
			int var_len, new_len;
			wchar_t *var_name;
			wchar_t * var_val;
			wchar_t * new_in;			
			
			while (in[stop_pos]==VARIABLE_EXPAND)			 
				stop_pos++;

			while( 1 )
			{
				if( !(in[stop_pos ]) )
					break;
				if( !( iswalnum( in[stop_pos] ) ||
					   (wcschr(L"_", in[stop_pos])!= 0)  ) )
					break;
			
				stop_pos++;
			}
			
/*			printf( "Stop for '%c'\n", in[stop_pos]);*/
			
			var_len = stop_pos - start_pos;
			
			if( !(var_name = malloc( sizeof(wchar_t)*(var_len+1) )))
			{
				error( OOM, L"Out of memory", -1 );				
				return 0;
			}
			
			wcsncpy( var_name, &in[start_pos], var_len );
			var_name[var_len]='\0';
/*			printf( "Variable name is %s, len is %d\n", var_name, var_len );*/
			
			var_val = expand_var( var_name );
			
			new_len = wcslen(in) - (stop_pos-start_pos+1) + wcslen( var_val) +1;

			if( !(new_in = malloc( sizeof(wchar_t)*new_len )))
			{
				error( OOM, L"Out of memory", -1 );				
				free(var_val);				
				free(var_name);				
				return 0;
			}
			
			wcsncpy( new_in, in, start_pos-1 );
			new_in[start_pos-1]='\0';
			wcscat( new_in, var_val );
			wcscat( new_in, &in[stop_pos] );

			i += wcslen(var_val)-1;

			free( var_name );
			//free( var_val );
			free(in);
			in=new_in;			
			
		}
		prev_char = c;		
	}
	*ptr = in;

	return 1;
}

/**
   Perform bracket expantion
*/
static int expand_brackets( wchar_t *in, int accept_incomplete, array_list *out )
{
	wchar_t *pos;
	int syntax_error=0;
	int bracket_count=0;	

	wchar_t *bracket_begin=0, *bracket_end=0;
	wchar_t *item_begin;
	int len1, len2, tot_len;
	
	for( pos=in; 
		 (!bracket_end) && (*pos); 
		 pos++ )
	{
		if( *pos == BRACKET_BEGIN )
		{
			if(( bracket_count == 0)&&(bracket_begin==0))
				bracket_begin = pos;
			
			bracket_count++;
		}
		else if( *pos == BRACKET_END )
		{
			bracket_count--;
			if( (bracket_count == 0) && (bracket_end == 0) )
				bracket_end = pos;
			
			if( bracket_count < 0 )
			{
				syntax_error = 1;
				break;
			}
		}
	}
	
	if( bracket_count > 0 )
	{
		if( !accept_incomplete )
			syntax_error = 1;
		else
		{
			wchar_t *woot = wcsdupcat( in, L"}" );			
			return expand_brackets( woot, 1, out );
		}
	}
	
	if( syntax_error )
	{
		error( SYNTAX_ERROR, L"Mismatched brackets", -1 );
		return 0;
	}
	
	if( bracket_begin == 0 )
	{
		al_push( out, in );
		return 1;
	}

	len1 = (bracket_begin-in);
	len2 = wcslen(bracket_end)-1;
	tot_len = len1+len2;
	item_begin = bracket_begin+1;
	for( pos=(bracket_begin+1); 1; pos++ )
	{
		if( bracket_count == 0 )
		{
			if( (*pos == ',') || (pos==bracket_end) )
			{
				wchar_t *whole_item;
				int item_len = pos-item_begin;
				
				whole_item = malloc( sizeof(wchar_t)*(tot_len + item_len + 1) );
				wcsncpy( whole_item, in, len1 );
				wcsncpy( whole_item+len1, item_begin, item_len );	
				wcscpy( whole_item+len1+item_len, bracket_end+1 );
				
				expand_brackets( whole_item, accept_incomplete, out );
				
				item_begin = pos+1;
				if( pos == bracket_end )
					break;
			}			
		}
		
		if( *pos == BRACKET_BEGIN )
		{
			bracket_count++;
		}
		
		if( *pos == BRACKET_END )
		{
			bracket_count--;
		}
	}
	free(in);	
	return 1;	
}

int expand_locate_subshell( wchar_t *in, 
							wchar_t **begin, 
							wchar_t **end,
							int allow_incomplete )
{
	wchar_t *pos;
	wchar_t prev=0;
	int syntax_error=0;
	int paran_count=0;	

	wchar_t *paran_begin=0, *paran_end=0;

	for( pos=in; *pos; pos++ )
	{
		if( prev != '\\' )
		{
			if( *pos == '(' )
			{
				if(( paran_count == 0)&&(paran_begin==0))
					paran_begin = pos;
				
				paran_count++;
			}
			
			if( *pos == ')' )
			{
				paran_count--;
				if( (paran_count == 0) && (paran_end == 0) )
				{
					paran_end = pos;
					break;
				}
				
				if( paran_count < 0 )
				{
					syntax_error = 1;
					break;
				}
			}
		}
		
		prev = *pos;
	}
	
	syntax_error |= (paran_count < 0 );
	syntax_error |= ((paran_count>0)&&(!allow_incomplete));
	
	if( syntax_error )
	{
		return -1;
	}
	
	if( paran_begin == 0 )
	{
		return 0;
	}

	*begin = paran_begin;
	*end = paran_count?in+wcslen(in):paran_end;
	
	return 1;

}

/**
   Perform subshell expantion
*/
static int expand_subshell( wchar_t *in, array_list *out )
{
	wchar_t *paran_begin=0, *paran_end=0;
	int len1, len2;
	wchar_t prev=0;
	wchar_t *subcmd;
	array_list sub_res, tail_expand;
	int i, j;
	wchar_t *item_begin;

	switch( expand_locate_subshell(in, 
								   &paran_begin,
								   &paran_end,
								   0 ) )
	{
		case -1:
			error( SYNTAX_ERROR, L"Mismatched parans", -1 );
			return 0;
		case 0:
			al_push( out, in );
			return 1;		
		case 1:
			
			break;
		
	}
	
	

	len1 = (paran_begin-in);
	len2 = wcslen(paran_end)-1;
	prev=0;
	item_begin = paran_begin+1;

	al_init( &sub_res );
	if( !(subcmd = malloc( sizeof(wchar_t)*(paran_end-paran_begin) )))
	{
		al_destroy( &sub_res );
		return 0;		
	}
	
	wcsncpy( subcmd, paran_begin+1, paran_end-paran_begin-1 );
	subcmd[ paran_end-paran_begin-1]=0;	
	
	if( !exec_subshell( subcmd, &sub_res) )
	{
		al_foreach( &sub_res, (void (*)(const void *))&free );
		al_destroy( &sub_res );
		free( subcmd );
		return 0;
	}	

	al_init( &tail_expand );
	expand_subshell( wcsdup(paran_end+1), &tail_expand );

	for( i=0; i<al_get_count( &sub_res ); i++ )
	{
		wchar_t *sub_item;
		sub_item = (wchar_t *)al_get( &sub_res, i );
		int item_len = wcslen( sub_item );

		for( j=0; j<al_get_count( &tail_expand ); j++ )
		{
			wchar_t *whole_item;
			wchar_t *tail_item = (wchar_t *)al_get( &tail_expand, j );
			int tail_len = wcslen( tail_item );
			
			whole_item = malloc( sizeof(wchar_t)*(len1 + item_len + tail_len + 1) );
			if( !whole_item )
			{
				wperror( L"malloc" );
				return 0;
			}
			
			wcsncpy( whole_item, in, len1 );
			wcsncpy( whole_item+len1, sub_item, item_len );	
			wcscpy(  whole_item+len1+item_len, tail_item );
			al_push( out, whole_item );			
		}
	}
	free(in);
	
	al_foreach( &sub_res, (void (*)(const void *))&free );
	al_destroy( &sub_res );
	
	al_foreach( &tail_expand, (void (*)(const void *))&free );
	al_destroy( &tail_expand );

	free( subcmd );
	return 1;	
}

/**
   Expand backslashed escapes and substitute them with their unescaped
   counterparts. Also optionally change the wildcards and tilde characters into
   constants which are defined to be outside of the valid character
   space, but still inside the valid space for a wchar_t. This
   assumes that a wchar_t is at least 32 bits long AND that the
   characterset is UCS4 or some other 31-bit character set.

   Since removing the escape sequences can never lengthen the string,
   the specified string is modified instead of allocating a new one.
*/
static wchar_t *expand_backslash_internal( wchar_t * in, int escape_special )
{
	int in_pos, out_pos, len = wcslen( in );
	int c;
	
	for( in_pos=0, out_pos=0; in_pos<len; out_pos++, in_pos++ )
	{
		c = in[in_pos];
		if( c == L'\\' )
		{
			switch( in[++in_pos] )
			{
				case L'\0':
					error( SYNTAX_ERROR, L"Unexpected end of string", -1 );
					break;

				case L'n':
					in[out_pos]=L'\n';
					break;

				case L'r':
					in[out_pos]=L'\r';
					break;

				case L't':
					in[out_pos]=L'\t';
					break;

				case L'b':
					in[out_pos]=L'\b';
					break;

				case L'e':
					in[out_pos]=L'\e';
					break;

				default:
					in[out_pos]=in[in_pos];
					break;
			}
		}
		else 
		{
			switch( in[in_pos]){
				case L'~':
					if( escape_special && (in_pos == 0) )
						in[out_pos]=HOME_DIRECTORY;
					else
						in[out_pos] = L'~';
					break;
				case L'%':
					if( escape_special && (in_pos == 0) )
						in[out_pos]=PROCESS_EXPAND;
					else
						in[out_pos] = L'%';
					break;
				case L'*':
					if( escape_special )
						in[out_pos]=ANY_STRING;
					else
						in[out_pos]=L'*';						
					break;
				case L'?':
					if( escape_special )
						in[out_pos]=ANY_CHAR;
					else
						in[out_pos]=L'?';						
					break;					
				case L'$':
					if( escape_special )
						in[out_pos]=VARIABLE_EXPAND;
					else
						in[out_pos]=L'$';						
					break;					
				case L'{':
					if( escape_special )
						in[out_pos]=BRACKET_BEGIN;
					else
						in[out_pos]=L'$';						
					break;					
				case L'}':
					if( escape_special )
						in[out_pos]=BRACKET_END;
					else
						in[out_pos]=L'$';						
					break;					
					
				default:
					in[out_pos] = in[in_pos];
					break;
			}
		}		
	}
	in[out_pos]=L'\0';
	return in;	
}

wchar_t *expand_backslash( wchar_t *in )
{
	return expand_backslash_internal( in, 0 );
}


/**
  Attempts tilde expansion. Of the string specified. If tilde
  expansion is performed, the argument is freed and a new string is
  allocated in its place. Horrible call signature. Should be
  altered. Fugly!
*/
static int tilde_expand( wchar_t **ptr )
{
	wchar_t *in = *ptr;
	
	if( in[0] == HOME_DIRECTORY )
	{
		int tilde_error = 0;
		wchar_t *home=0;
		wchar_t *new_in;
		wchar_t *old_in;
		
//		fwprintf( stderr, L"Tilde expand ~%ls\n", (*ptr)+1 );
		if( in[1] == '/' || in[1] == '\0' )
		{
			/* Current users home directory */
			struct passwd *userinfo;
			
			while( 1 )
			{
				userinfo=getpwuid( getuid() );
				if( userinfo )
					break;
				else if( errno != EINTR )
				{
					wperror( L"getpwuid" );
				}				
			}

			if( userinfo )
			{	
				home = str2wcs(userinfo->pw_dir);
			}
			else
				home = wcsdup(L"");

			old_in = &in[1];
		}
		else
		{
			/* Some other users home directory */
			wchar_t *name;
			wchar_t *name_end = wcschr( in, L'/' );
			if( name_end == 0 )
			{	
				name_end = in+wcslen( in );
			}
			name = wcsndup( in+1, name_end-in-1 );
			old_in = name_end;
			
			char *name_str = wcs2str( name );			

			struct passwd *userinfo = 
				getpwnam( name_str );
			
			if( userinfo == 0 )
			{
				tilde_error = 1;
				*in = L'~';
			}
			else
			{
				home = str2wcs(userinfo->pw_dir);
			}
			
			free( name_str );
			free(name);
		}
		
		if( !tilde_error )
		{
			new_in = wcsdupcat( home, old_in ); 
			free( in );
			in = new_in;
			free(home);
			*ptr = in;
		}		
		
	}
	return 1;
} 

wchar_t *expand_tilde(wchar_t *in)
{
	if( in[0] == L'~' )
	{
		in[0] = HOME_DIRECTORY;
		tilde_expand( &in );
		return in;
	}
	return in;	
}


/**
  Check if the specified string contains wildcards
*/
static int has_wildcard( wchar_t *str )
{
	wchar_t prev=0;
	for( ; *str; str++ )
	{
		if( ( (*str == ANY_CHAR ) || (*str == ANY_STRING) ) && (prev != L'\\') )
			return 1;
		prev = *str;
	}
	return 0;
}

/**
  The real expansion function. All other expantion  functions are wrappers to this one.
*/
static int expand_string_internal( wchar_t *str,
								   array_list *end_out, 
								   int subshell_flag,
								   int accept_incomplete )
{
	array_list list1, list2;
	array_list *in, *out;
	
	int i;
	int subshell_ok = 1;
	
	al_init( &list1 );
	al_init( &list2 );

	if( subshell_flag )
	{
		subshell_ok = expand_subshell( str, &list1 );
	}
	else
	{
		wchar_t *pos = str;
		
		while( 1 )
		{
			pos = wcschr( pos, L'(' );
			if( pos == 0 )
				break;
			
			if( (pos == str) || ( *(pos-1) != L'\\' ) )
			{
				error( SUBSHELL_ERROR, L"Subshells not allowed", -1 );
				al_destroy( &list1 );
				al_destroy( &list2 );
				return 0;
			}
			pos++;
		}		
		al_push( &list1, str );
	}

	if( !subshell_ok )
	{
		al_destroy( &list1 );
		return 0;
	}
	else
	{
		in = &list1;
		out = &list2;

		for( i=0; i<al_get_count( in ); i++ )
		{
			wchar_t *next = expand_backslash_internal((wchar_t *)al_get( in, 
																		 i ), 
													  1);
			
			if(!expand_variables( &next ))
			{
				al_destroy( in );
				al_destroy( out );
				return 0;	
			}
			
			if( !expand_brackets( next, accept_incomplete, out ))
			{
				al_destroy( in );
				al_destroy( out );
				return 0;
			}
		}
		al_truncate( in, 0 );

		in = &list2;
		out = &list1;
		
		for( i=0; i<al_get_count( in ); i++ )
		{
			wchar_t *next = (wchar_t *)al_get( &list2, i );		
						
			if(!tilde_expand( &next ))
			{
				al_destroy( in );
				al_destroy( out );
				return 0;
			}
		
			if( accept_incomplete )
			{
				if( *next == PROCESS_EXPAND )
				{
					expand_pid( next, accept_incomplete, end_out );
					al_destroy( in );
					al_destroy( out );
					return 1;
				}
				else
					al_push( out, next );
			}
			else
			{
				if( !expand_pid( next, accept_incomplete, out ) )
				{
					al_destroy( in );
					al_destroy( out );
					return 0;
				}
			}
		}
		al_truncate( in, 0 );
		
		in = &list1;
		out = &list2;		

		
		for( i=0; i<al_get_count( in ); i++ )
		{
			wchar_t *next = (wchar_t *)al_get( in, i );
			int wc_res;

			if( accept_incomplete || has_wildcard( next ) )
			{
				
				if( next[0] == '/' )
				{
					wc_res = wildcard_expand( &next[1], L"/",accept_incomplete, out );
				}
				else
				{
					wc_res = wildcard_expand( next, L"", accept_incomplete, out );
				}
				free( next );
				switch( wc_res )
				{
					case 0:
						if( !accept_incomplete )
						{	
							break;
						}
					case 1:
						sort_list( out );
						al_push_all( end_out, out );
						al_truncate( out, 0 );						
						break;
						
					default:
						fwprintf( stderr, L"error\n" );
						/*al_destroy( &list1 );*/
						/*al_destroy( &list2 );*/
						/*return 0;*/
				}			
			}
			else
			{
				if( !accept_incomplete)
					al_push( end_out, next );
			}		
		}
		al_destroy( in );
		al_destroy( out );
	}
	return 1;
}



int expand_string( wchar_t *str, 
				   array_list *out,
				   int flag )
{
	return expand_string_internal( str, out, flag, 0 );
}

wchar_t *expand_one( wchar_t *string, int flag )
{
	array_list l;
	int res;
	wchar_t *one;
	
	al_init( &l );
	res = expand_string_internal( string, &l, flag, 0 );
	if( !res )
	{
		free( string );
		one = 0;
	}
	else
	{
		if( al_get_count( &l ) != 1 )
		{
			one=0;
		}
		else
		{
			one = (wchar_t *)al_get( &l, 0 );
			al_set( &l, 0, 0 );
		}
	}
	
	al_foreach( &l, (void(*)(const void *))&free );
	al_destroy( &l );
	return one;
}

int expand_file_completion( wchar_t *str,
					   array_list *out )
{
	return expand_string_internal( str, out, 0, 1 );
}

