/**\file expand.h
   Prototypes for string expantion functions. These functions perform several kinds of parameter expantion.
*/

/**
  Perform various forms of expansion on in, such as tilde expansion
  (~USER becomes the users home directory), variable expansion
  ($VAR_NAME becomes the value of the environment variable VAR_NAME),
  subshell expantion and wildcard expansion. The results are inserted
  into the list out.

  If the parameter does not need expansion, it is copied into the list
  out. If expansion is performed, the original parameter is freed and
  newly allocated strings are inserted into the list out.

  \param in The parameter to expand
  \param subshell_ok True if subshells should be executed, false if subshells should cause the expantion to fail.
  \param out The list to which the result will be appended.
*/
int expand_string( wchar_t *in, array_list *out, int subshell_ok );

/**
  expand_one is identical to expand_string, except it will fail if in
  expands to more than one string. This is used for expanding command names.

  \param in The parameter to expand
  \param subshell_ok True if subshells should be executed, false if subshells should cause the expantion to fail.
  \return The expanded parameter, or 0 on failiure
*/
wchar_t *expand_one( wchar_t *in, int subshell_ok );

/**
  expand_file_completion is used by tab-completion to find file
  completions. It performs an expantion like expand_string, but
  instead of returning the expanded string, returns all possible
  completions which result in an existing file. Subshells are not
  allowed in the parameter.
  
  \param in The parameter to expand
  \param out The list to which the result will be appended.
*/
int expand_file_completion( wchar_t *in, array_list *out );

/**
  Replace special characters with escape sequences. Newline is
  replaced with \n, etc. 

  \param in The string to be escaped
  \param escape_all Whether all characters wich hold special meaning in fish (Pipe, semicolon, etc,) should be escaped, or only unprintable characters
  \return The escaped string, or 0 if there is not enough memory
*/
wchar_t *expand_escape( const wchar_t *in, int escape_all );


/**
   Perform tilde expantion and nothing else on the specified string.

   If tilde expantion is needed, the original string is ferred and a
   new string, allocated using malloc, is returned.
*/
wchar_t *expand_tilde(wchar_t *in);

/**
   Locate the last subshell in the specified string.
   
   \param in the string to search for subshells
   \param begin the starting paranthesis of the subshell
   \param end the ending paranthesis of the subshell
   \param allow_incomplete expand for tabcompletion
   \return -1 on syntax error, 0 if no subshells exist and 1 on sucess
*/
int expand_locate_subshell( wchar_t *in, 
							wchar_t **begin, 
							wchar_t **end,
							int allow_incomplete );

/**
   Expand backslashed escapes and substitute them with their unescaped
   counterparts. 

   Since removing the escape sequences can never lengthen the string,
   the specified string is modified instead of allocating a new one.
*/
wchar_t *expand_backslash( wchar_t * in );


