/** \file history.c
  History functions, part of the user interface.
*/
#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <errno.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#include "config.h"
#include "util.h"
#include "wutil.h"
#include "history.h"
#include "common.h"
#include "reader.h"
#include "env.h"
#include "sanity.h"


/** Maximum number of items in history */
#define HIST_MAX 8192

/** Name of history file */
#define HISTORY_FILE L"/.fish_history"

/*
  The history is implemented using a linked list. Searches are done
  using linear search. 
*/

static ll_node /** Last item in history */ *history_last=0, /** Current search position*/ *history_current =0;

/**
   We only try to save the part of the history that was written in
   this session. This pointer keeps track of where the cutoff is.

   The reason for wanting to do this is so that when two concurrent
   fish sessions have exited, the latter one to exit won't push the
   added history items of the prior session to the top of the history.
*/
static ll_node *last_loaded=0;

/**
  past_end is a kludge. It is set when the current history item is
  unset, i.e. a new entry is beeing typed.
*/
static int past_end =1;

/**
   Items in history
*/
static int history_count=0;

/**
   Load history from ~/.fish_history
*/
static void history_load()
{
	wchar_t *fn;
	wchar_t *buff=0;
	int buff_len=0;
	FILE *in_stream;
	hash_table used;
	hash_init( &used, 
			   &hash_wcs_func,
			   &hash_wcs_cmp );
	
	fn = wcsdupcat( env_get(L"HOME"), HISTORY_FILE );
	in_stream = wfopen( fn, "r" );
	if( in_stream != 0 )
	{
		while( !feof( in_stream ) )
		{
			int buff_read = fgetws2( &buff, &buff_len, in_stream );
			if( buff_read == -1 )
			{
				wperror( L"fish" );
				fwprintf( stderr, L"Error reading history file\n" );
				fclose( in_stream );
				free( fn );
				free( buff );
				return;
			}

			/*
			  We do not call history_add here, since that would make
			  history_load() use quadratic time, and may be
			  unacceptably on slow systems with a long history file.

			  Use a hashtable to check for duplicates instead.
			*/
			if( !hash_get( &used,
						   buff ) )
			{
				history_count++;		
				
				history_current = malloc( sizeof( ll_node ) );
				if( !history_current )
				{
					fwprintf( stderr, L"fish: Out of memory in history.c\n" );
					return;
				}
				
				history_current->data=wcsdup( buff );

				hash_put( &used,
						  history_current->data, 
						  L"" );
				
				history_current->next=0;
				history_current->prev = history_last;
				if( history_last != 0 )
				{
					history_last->next = history_current;
				}
				history_last = history_current;
			}
		}
		
		fclose( in_stream );
	}
	else
	{
		wperror( L"fopen" );
	}	
	hash_destroy( &used );
	
	free( buff );
	free( fn );
	last_loaded = history_last;
}

void history_init()
{
	history_load();
}

/**
   Print history node 
*/
static void history_save_node( ll_node *n, FILE *out )
{
	if( n==0 )
		return;
	history_save_node( n->prev, out );
	fwprintf(out, L"%ls\n", (wchar_t *)(n->data) );
}

/**
   Merge the process history with the history on file, write to
   disc. The merging operation si done so that two cuncurrently
   running shells wont erase each others history.
*/
static void history_save()
{
	wchar_t *fn;
	FILE *out_stream;
	/* First we save this sessions history in local variables */
	ll_node *real_pos = history_last, *real_first = last_loaded;
	if( !real_first )
	{
		real_first = history_current;
		while( real_first->prev )
			real_first = real_first->prev;
	}
	
	/* Then we load the history from file into the global pointers */
	history_last=history_current=last_loaded=0;
	history_count=0;
	past_end=1;

	history_load();
	if( real_pos != 0 )
	{
		/* 
		   Rewind the session history to the first item which was
		   added in this session
		*/
		while( (real_pos->prev != 0) && (real_pos->prev != real_first) )
		{
			real_pos = real_pos->prev;
		}
		

		/* Free old history entries */
		ll_node *kill_node = real_pos->prev;
		while( kill_node != 0 )
		{
			ll_node *tmp = kill_node;
			free( kill_node->data );
			kill_node = kill_node->prev;
			free( tmp );		
		}	

		/* 
		   Add all the history entries from this session to the global
		   history, free the old version
		*/
		while( real_pos != 0 )
		{
			ll_node *next = real_pos->next;
			history_add( (wchar_t *)real_pos->data );
			
			free( real_pos->data );
			free( real_pos );
			real_pos = next;
		}
	}

	/* Save the global history */
	fn = wcsdupcat( env_get(L"HOME"), HISTORY_FILE );
	
	out_stream = wfopen( fn, "w" );
	if( out_stream != 0 )
	{
		history_save_node( history_last, out_stream );
		fclose( out_stream );
	}
	else
	{
		wperror( L"fopen" );
	}
	free( fn );	
}

void history_destroy()
{
	history_save();	
	history_current = history_last;
	while( history_current != 0 )
	{
		ll_node *tmp = history_current;
		free( history_current->data );
		history_current = history_current->prev;
		free( tmp );		
	}	
}

/**
   Internal search function
*/
static ll_node *history_find( ll_node *n, const wchar_t *s )
{
	if( n == 0 )
		return 0;
	
	if( wcscmp( s, (wchar_t *)n->data ) == 0 )
	{
		return n;		
	}
	else
		return history_find( n->prev, s );
}


void history_add( const wchar_t *str )
{
	ll_node *old_node;

	if( wcslen( str ) == 0 )
		return;
	
	past_end=1;

	old_node = history_find( history_last, str );

	if( old_node == 0 )
	{
		history_count++;		
		history_current = malloc( sizeof( ll_node ) );
		history_current->data=wcsdup( str );
	}
	else
	{
		if( old_node == last_loaded )
		{
			last_loaded = last_loaded->prev;
		}
	
		history_current = old_node;
		if( old_node == history_last )
		{
			return;
		}

		if( old_node->prev != 0 )
			old_node->prev->next = old_node->next;
		if( old_node->next != 0 )
			old_node->next->prev = old_node->prev;
	}
	history_current->next=0;
	history_current->prev = history_last;
	if( history_last != 0 )
	{
		history_last->next = history_current;
	}
	history_last = history_current;
}

const wchar_t *history_prev_match( const wchar_t *str )
{
	if( history_current == 0 )
		return str;
	
	if( history_current->prev == 0 )
	{
		if( wcsncmp( (wchar_t *)history_current->data, str, wcslen(str) )==0 )
			return (wchar_t *)history_current->data;
		else
			return str;
	}
	if( past_end )
		past_end = 0;
	else
		history_current = history_current->prev;

	if( wcsncmp( (wchar_t *)history_current->data, str, wcslen(str) )==0 )
		return (wchar_t *)history_current->data;
	else
		return history_prev_match( str );
}


const wchar_t *history_next_match( const wchar_t *str)
{
	if( history_current == 0 )
		return str;

	if( history_current->next == 0 )
	{
		past_end = 1;
		return str;
	}
	history_current = history_current->next;
	
	if( wcsncmp( (wchar_t *)history_current->data, str, wcslen(str) )==0 )
		return (wchar_t *)history_current->data;
	else
		return history_next_match( str );	
}

void history_reset()
{
	history_current = history_last;
}

void history_sanity_check()
{	
	if( history_current != 0 )
	{
		int i;
		int history_ok = 1;
		int found_current = 0;
		
		validate_pointer( history_last, L"History root", 1);
		
		ll_node *tmp = history_last;
		for( i=0; i<history_count; i++ )
		{
			found_current |= tmp == history_current;

			if( tmp == 0 )
			{
				history_ok = 0;
				fwprintf( stderr, L"fish: History items missing\n" );
				break;
			}

			if( (tmp->prev != 0) && (tmp->prev->next != tmp ) )
			{
				history_ok = 0;
				fwprintf( stderr, L"fish: History items order is inconsistent\n" );
				break;
			}
			
			validate_pointer( tmp->data, L"History data", 1);
			if( tmp->data == 0 )
			{
				history_ok = 0;
				fwprintf( stderr, L"fish: Empty history item\n" );
				break;
			}
			validate_pointer( tmp->prev, L"History node", 1);
			tmp = tmp->prev;
		}
		if( tmp != 0 )
		{
			history_ok = 0;
			fwprintf( stderr, L"fish: History list too long\n" );
		}

		if( (i!= history_count )|| (!found_current))
		{
			fwprintf( stderr, L"fish: No history item selected\n" );
			history_ok=0;
		}
		
		if( !history_ok )
		{
			sanity_lose();			
		}

	}	
}

