/** \file reader.c
	
Functions for reading data from stdin and passing to the
parser. If stdin is a keyboard, it supplies a killring, history,
syntax highlighting, tab-completion and various other features.
*/

#include "config.h"

#include <wchar.h>

/* 
   On my system, a weird compiler error is issued if this file is
   included further down. Haven't had the time to check what really
   happens...
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>

#if HAVE_NCURSES_H
#include <ncurses.h>
#else
#include <curses.h>
#endif


#if HAVE_TERMIO_H
#include <termio.h>
#endif

#include <term.h>
#include <signal.h>
#include <fcntl.h>
#include <dirent.h>
#include <time.h>

#include "util.h"
#include "wutil.h"
#include "highlight.h"
#include "reader.h"
#include "proc.h"
#include "parser.h"
#include "complete.h"
#include "history.h"
#include "common.h"
#include "sanity.h"
#include "env.h"
#include "exec.h"
#include "expand.h"
#include "tokenizer.h"
#include "kill.h"
#include "input.h"

/**
   Color code for set_color. Does not update the color.
*/

#define FISH_COLOR_IGNORE -1

/**
   Color code for set_color. Sets the default color.
*/
#define FISH_COLOR_RESET -2



/** 
	Maximum length of prefix string when printing completion
	list. Longer prefixes will be ellipsized.
*/
#define PREFIX_MAX_LEN 16





/**
   The current size of the input buffers
*/
static size_t buff_alloc_sz = 0;

/**
   Buff contains the curreent command that is edited.
*/
static wchar_t *buff=0;

/**
   prefix_buff contains the prefix that is used in searching the
   history.
*/
static wchar_t *prefix_buff=0;
/**
   Color is the syntax highlighting for buff.  The format is that
   color[i] is the classification (according to the enum in
   highlight.h) of buff[i].
*/
static int *color=0;

/**
   New color buffer, used for syntax highlighting.
*/
static int *new_color=0;

/**
   The current position of the cursor in buff.
*/
static int buff_pos=0;

/**
   Length of the command in buff. (Not the length of buff itself)
*/
static int buff_len=0;

/** When this is true, the shell will exit */
static int end_loop = 0;

/**
   This struct should be continually updated by signals as the term resizes, and as such always contain the correct current size.
*/
static struct winsize termsize;

/**
   Returns the line number of the line that is currently parsed
*/
static int current_lineno;

/**
   The list containting names of files that are beeing parsed
*/
static array_list current_filename;

/**
   These status buffers are used to check if any output has occured
   other than from fish's main loop, in which case we need to redraw.
*/
static struct stat prev_buff_1, prev_buff_2, post_buff_1, post_buff_2;



/**
   List containing strings which make up the prompt
*/
static array_list prompt_list;

/**
   Stores the previous termios mode so we can reset the modes when
   we execute programs and when the shell exits.
*/	
static struct termios saved_modes;


/**
   Store the pid of the parent process, so the exit function knows whether it should reset the terminal or not.
*/
static pid_t original_pid;

/**
   Interupted flag. Set to 1 when the user presses \^C. 
*/
static int interupted;

/*
  Prototypes for a bunch of functions 
*/

static void reader_save_status();
static void reader_check_status();


/**
   Test if there is space between the time fields of struct stat to use
   for sub second information.
*/
static int room_for_usec(struct stat *st)
{
	int res = ((&(st->st_atime) + 2) == &(st->st_mtime) && 
			   (&(st->st_atime) + 4) == &(st->st_ctime));
	return res;
}

int reader_get_width()
{
	return termsize.ws_col;
}


int reader_get_height()
{
	return termsize.ws_row;
}	

int reader_current_lineno()
{
	return current_lineno;
}


wchar_t *reader_current_filename()
{
	return (wchar_t *)al_peek( &current_filename );
}


void reader_push_current_filename( wchar_t *fn )
{
	al_push( &current_filename, fn );
}


wchar_t *reader_pop_current_filename()
{
	return (wchar_t *)al_pop( &current_filename );
}



/**
   Make sure buffers are large enough to hold current data.
*/
static int check_size()
{
	if( buff_alloc_sz < buff_len + 2 )
	{
		buff_alloc_sz = maxi( 128, buff_len*2 );

		buff = realloc( buff, sizeof(wchar_t)*buff_alloc_sz);
		prefix_buff = realloc( prefix_buff, sizeof(wchar_t)*buff_alloc_sz);
		color = realloc( color, sizeof(int)*buff_alloc_sz);
		new_color = realloc( new_color, sizeof(int)*buff_alloc_sz);
		
		if( buff==0 || prefix_buff==0 || color==0 || new_color == 0 )
		{
			wperror( L"realloc" );
			fwprintf( stderr, L"Out of memory while allocating read buffer\n" );
			sanity_lose();
			return 0;
		}		
	}
	
	return 1;
}

/**
   Compare two completions, ignoring their description.
*/
static int fldcmp( wchar_t *a, wchar_t *b )
{
	while( 1 )
	{
		if( *a != *b )
			return *a-*b;
		if( ( (*a == COMPLETE_SEP) || (*a == L'\0') ) && 
			( (*b == COMPLETE_SEP) || (*b == L'\0') ) )
			return 0;
		a++;
		b++;
	}
	
}

/**
   Remove any duplicate completions in the list.
*/
static void remove_duplicates( array_list *l )
{
	int in, out;
	wchar_t *prev;
	if( al_get_count( l ) == 0 )
		return;
	
	prev = (wchar_t *)al_get( l, 0 );
	for( in=1, out=1; in < al_get_count( l ); in++ )
	{
		wchar_t *curr = (wchar_t *)al_get( l, in );
		if( fldcmp( prev, curr )==0 )
		{
			free( curr );
		}
		else
		{
			al_set( l, out++, curr );
			prev = curr;
		}		
	}
	al_truncate( l, out );
}


/**
   Write a wide character to FD 1
*/
static int writech( wint_t ch )
{
	static mbstate_t out_state;
	char buff[MB_CUR_MAX];
	size_t bytes = wcrtomb( buff, ch, &out_state );
	int err;
	
	while( (err =write( 1, buff, bytes ) ) )
	{
		if( err >= 0 )
			break;

		if( errno == EINTR )
			continue;

		wperror( L"write" );
		return 1;
	}
	
	return 0;
}

/**
   Write a wide character string to FD 1.   
*/
static int writestr( const wchar_t *str )
{
	while( *str != 0 )
		writech( *(str++) );
	
	return 0;
}

/**
   Escape and write a string to fd 1
*/
static int write_escaped_str( const wchar_t *str, int max_len )
{

	wchar_t *out = expand_escape( str, 1 );
	int i;
	int len = wcslen( out );
	int written=0;
	
	if( max_len && (max_len < len))
	{
		for( i=0; i<max_len-3; i++ )
		{
			writech( out[i] );
		}
		for( i=0; i<mini( 3, max_len ); i++ )
			writech( L'.' );
		written = max_len;
		
	}
	else
	{
		written = len;
		writestr( out );
	}
	
	free( out );
	return written;
}

/**
   We cant use putchar as the output function for putp, since we
   ourselves use FD based IO, which should not be mixed to liberally
   with stream IO. So we provide this putchar workalike.
*/

static int writeb( tputs_arg_t b )
{
	write( 1, &b, 1 );
	return 0;
}

/**
   Write a char * narrow string to FD 1, needed for the terminfo
   strings.
*/
static int writembs( char *str )
{
	tputs(str,1,&writeb);
	return 0;
}

/**
   parm_ich seems to often be undefined, so we use this
   workalike. Writes the specified number of spaces.
*/
static int writespace( int c )
{
	write( 1, "        ", mini(c,8) );
	if( c>8)
	{
		writespace( c-8);
	}
	return 0;	
}


/**
   Sets the fg and bg color. May be called as often as you like, since
   if the new color is the same as the previous, nothing will be
   written. Negative values for set_color will also be ignored. Since
   the terminfo string this function emits can potentially cause the
   screen to flicker, the function takes care to write as little as
   possible.

   Possible values for color are any form the FISH_COLOR_* enum,
   FISH_COLOR_IGNORE and FISH_COLOR_RESET. FISH_COLOR_IGNORE will
   leave the color unchanged, and FISH_COLOR_RESET will perform an
   exit_attribute_mode, even if set_color thinks it is already in
   FISH_COLOR_NORMAL mode.

   In order to set the color to normal, three terminfo strings may
   have to be written.

   - First a string to set the color, such as set_a_foreground. This
   is needed bacuase otherwise the previous strings colors might be
   removed as well.

   - After that we write the exit_attribute_mode string to reset all
   color attributes.

   - Lastly we may need to write set_a_background or set_a_foreground
   to set the other half of the color pair to what it should be.

   \param c Foreground color.
   \param c2 Background color.
*/


static void set_color( int c, int c2 )
{
	static int last_color = FISH_COLOR_NORMAL, last_color2=FISH_COLOR_NORMAL;

	char *fg = 0, *bg=0;
	
	if( (set_a_foreground != 0) && (strlen( set_a_foreground) != 0 ) )
	{
		fg = set_a_foreground;
		bg = set_a_background;
	}
	else if( (set_foreground != 0) && (strlen( set_foreground) != 0 ) )
	{
		fg = set_foreground;
		bg = set_background;
	}

	if( (c == FISH_COLOR_RESET) || (c2 == FISH_COLOR_RESET))
	{
		c = c2 = FISH_COLOR_NORMAL;
		if( fg )
			writembs( tparm( set_a_foreground, 0 ) );		
		writembs( exit_attribute_mode );	
		return;
	}

	if( last_color != c )
	{
		if( c==FISH_COLOR_NORMAL )
		{
			if( fg )
				writembs( tparm( fg, 0 ) );		
			writembs( exit_attribute_mode );	
			
			last_color2 = FISH_COLOR_NORMAL;
			last_color = c;
		}
		else if( ( c >= 0) && ( c < FISH_COLOR_NORMAL ) )
		{
			if( fg )
				writembs( tparm( set_a_foreground, c ) );
			else if( (enter_bold_mode != 0) && (strlen(enter_bold_mode) > 0))
			{
				writembs( enter_bold_mode );
			}
			
			last_color = c;
		}
	}

	if( last_color2 != c2 )
	{
		if( c2 == FISH_COLOR_NORMAL )
		{
			if( bg )
				writembs( tparm( set_a_background, 0 ) );
			writembs(exit_attribute_mode);
			if(( last_color != FISH_COLOR_NORMAL ) && fg )
				writembs(tparm( set_a_foreground, last_color ));
			last_color2 = c2;
		}
		else if ((c2 >= 0 ) &&(c2 < FISH_COLOR_NORMAL))
		{
			if( bg )
				writembs(tparm( set_a_background, c2));
			last_color2 = c2;
		}
	}
}

static void set_color_translated( int c )
{
	
	set_color( highlight_get_color(c & 0xff), highlight_get_color( (c>>8)&0xff) );	
}

int reader_interupted()
{
	int res=interupted;
	if( res )
		interupted=0;
	return res;
}

void reader_write_title()
{
	char *n = ttyname( STDIN_FILENO );
	wchar_t *title;
	array_list l;

	if( strstr( n, "tty" ) )
		return;

	title = env_get( L"FISH_TITLE" );

	if( title == 0 )
	{
		title = L"echo $_ \" \"; pwd";
	}

	if( wcslen(title) ==0)
		return;
	
	al_init( &l );
	
	if( exec_subshell( title, &l ) )
	{
		int i;
		writestr(L"\e]2;");
		for( i=0; i<al_get_count( &l); i++ )
		{
			writestr( (wchar_t *)al_get( &l, i ) );			
		}
		writestr(L"\7");
	}		
	al_foreach( &l, (void (*)(const void *))&free );
	al_destroy( &l );
	set_color( FISH_COLOR_RESET, FISH_COLOR_RESET );
}

/**
   Write the prompt to screen.

   \param exec_prompt If true, the prompt should be recalculated by executing FISH_PROMPT in a subshell.
*/
static void write_prompt( int exec_prompt )
{
	set_color( FISH_COLOR_NORMAL, FISH_COLOR_NORMAL );
	if( exec_prompt != 0 )
	{	
		wchar_t *prompt = env_get( L"FISH_PROMPT" );
		al_foreach( &prompt_list, (void (*)(const void *))&free );
		al_truncate( &prompt_list, 0 );	
				
		if( prompt == 0 )
		{
			prompt = L"whoami; echo @; hostname|cut -d . -f 1; echo \" \"; set_color green; pwd; set_color normal; echo '> ';";
		}
		
		if( !exec_subshell( prompt, &prompt_list ) )
		{
			/* If executing the prompt fails, make sure we at least dont print any junk */
			al_foreach( &prompt_list, (void (*)(const void *))&free );
			al_destroy( &prompt_list );	
			al_init( &prompt_list );
		}
		exec_prompt = 0;
		reader_write_title();
	}
	
	int i;
	for( i=0; i<al_get_count( &prompt_list); i++ )
	{
		writestr( (wchar_t *)al_get( &prompt_list, i ) );			
	}
	set_color( FISH_COLOR_RESET, FISH_COLOR_RESET );

}

/**
   Write the whole command line (but not the prompt) to the screen. Do
   not set the cursor correctly afterwards.
*/
static void write_cmdline()
{
	int i, steps;
	for( i=0; i<buff_len; i++ )
	{
		set_color_translated( color[i] );
		
		writech( buff[i] );
	}
	steps = buff_len-buff_pos;
/*	writembs(tparm( parm_left_cursor, steps)); */
}

/**
   perm_left_cursor and parm_right_cursor don't seem to be defined as
   often as cursor_left and cursor_right, so we use this workalike.
*/
static void move_cursor( int steps )
{
	int i;
	
	if( steps < 0 ){
		for( i=0; i>steps; i--)
			writembs(cursor_left);
	}
	else
		for( i=0; i<steps; i++)
			writembs(cursor_right);
}


void reader_init()
{
	al_init( &current_filename);
}


void reader_destroy()
{
	al_destroy( &current_filename);
}



void do_exit()
{
	end_loop=1;
}

void repaint()
{
	set_color( FISH_COLOR_RESET, FISH_COLOR_RESET );
	writech('\r');
	writembs(clr_eol);				
	write_prompt(0);
	highlight( buff, color, buff_pos, 0 );
	write_cmdline();
	move_cursor(buff_pos-buff_len);
	set_color( FISH_COLOR_NORMAL, -1 );
	reader_save_status();
}

void check_colors()
{
	highlight( buff, new_color, buff_pos, 0 );
	if( memcmp( new_color, color, sizeof(int)*buff_len )!=0 ) 	
	{
		repaint();
	}
}



/**
   Stat stdout and stderr and save result.

   This should be done before calling a function that may cause output.
*/

static void reader_save_status()
{

#if (defined(__FreeBSD__) ||defined(__NetBSD__)) 
	/*
	  This futimes call tries to trick the system into using st_mtime
	  as a tampering flag. This of course only works on systems where
	  futimes is defined, but it should make the status saving stuff
	  failsafe.
	*/
	struct timeval t=
		{
			time(0)-1,
			0
		}
	;
	
	if( futimes( 1, &t ) || futimes( 2, &t ) )
	{
		wperror( L"futimes" );
	}
#endif

	fstat( 1, &prev_buff_1 );
	fstat( 2, &prev_buff_2 );
}

/**
   Stat stdout and stderr and compare result to previous result in
   reader_save_status. Repaint if modification time has changed.

   Unfortunatly, for some reason this call seems to give a lot of
   false positives, at least undeer Linux.
*/

static void reader_check_status()
{
	fflush( stdout );
	fflush( stderr );
	
	fstat( 1, &post_buff_1 );
	fstat( 2, &post_buff_2 );

	int changed = ( prev_buff_1.st_mtime != post_buff_1.st_mtime ) || 
		( prev_buff_2.st_mtime != post_buff_2.st_mtime );
	
	if (room_for_usec( &post_buff_1))
	{
		changed = changed || ( (&prev_buff_1.st_mtime)[1] != (&post_buff_1.st_mtime)[1] ) ||
			( (&prev_buff_2.st_mtime)[1] != (&post_buff_2.st_mtime)[1] );
	}

	//fwprintf( stderr, L"from %d %d to %d %d\n", prev_buff_1.st_mtime, (&prev_buff_1.st_mtime)[1], post_buff_1.st_mtime, (&post_buff_1.st_mtime)[1] );
	
	if( changed )
	{
		repaint();
		set_color( FISH_COLOR_RESET, FISH_COLOR_RESET );
	}
}

/**
   Remove the previous character in the character buffer and on the
   screen using syntac highlighting, etc.
*/
static void remove_backward()
{
	if( buff_pos <= 0 )
		return;

	if( buff_pos < buff_len )
	{
		memmove( &buff[buff_pos-1], &buff[buff_pos], sizeof(wchar_t)*(buff_len-buff_pos+1) );
		memmove( &color[buff_pos-1], &color[buff_pos], sizeof(wchar_t)*(buff_len-buff_pos+1) );
	}
	buff_pos--;
	buff_len--;
	buff[buff_len]='\0';
	wcscpy(prefix_buff,buff);
	writembs(cursor_left);

	history_reset();
	
	highlight( buff, new_color, buff_pos, 0 );
	if(( memcmp( new_color, color, sizeof(int)*buff_len )==0 ) &&
	   (delete_character != 0) )
	{
		if( enter_delete_mode != 0 )
			writembs(enter_delete_mode);
		writembs(delete_character);
		if( exit_delete_mode != 0 )
			writembs(exit_delete_mode);
	}
	else
	{
		memcpy( color, new_color, sizeof(int) * buff_len );
		repaint();
	}
}

/**
   Remove the current character in the character buffer and on the
   screen using syntac highlighting, etc.
*/
static void remove_forward()
{
	if( buff_pos >= buff_len )
		return;
	
	buff_pos++;
	writembs( cursor_right );

	remove_backward();
}

/**
   Insert the character into the command line buffer and print it to
   the screen using syntax highlighting, etc.
*/
static int insert_char( int c )
{

	if( !check_size() )
		return 0;
	
	/* Insert space for extra character at the right position */
	if( buff_pos < buff_len )
	{
		memmove( &buff[buff_pos+1], &buff[buff_pos], sizeof(wchar_t)*(buff_len-buff_pos) );
		memmove( &color[buff_pos+1], &color[buff_pos], sizeof(int)*(buff_len-buff_pos) );
	}
	/* Set character */
	buff[buff_pos]=c;

    /* Update lengths, etc */
	buff_pos++;
	buff_len++;
	buff[buff_len]='\0';
	
	/* Syntax highlight */
	highlight( buff, new_color, buff_pos, 0 );
	color[buff_pos-1] = new_color[buff_pos-1];
	/* Check if the coloring has changed */
	if(( memcmp( new_color, color, sizeof(int)*buff_len )==0 ) &&
	   (insert_character || (buff_pos == buff_len) || enter_insert_mode) )
	{
		/* Colors look ok, so we set the right color and insert a character */
		set_color_translated( color[buff_pos-1] );
		if( buff_pos < buff_len )
		{
			if( enter_insert_mode != 0 )
				writembs(enter_insert_mode);
			else
				writembs(insert_character);
			writech(c);
			if( insert_padding != 0 )
				writembs(insert_padding);
			if( exit_insert_mode != 0 )
				writembs(exit_insert_mode);
		}
		else
			writech(c);
		set_color( FISH_COLOR_NORMAL, -1 );
	}
	else
	{
		/* Nope, colors are off, so we repaint the entire command line */
		memcpy( color, new_color, sizeof(int) * buff_len );
		repaint();
	}
	wcscpy(prefix_buff,buff);
	return 1;
}

/**
   Insert the characters of the string into the command line buffer
   and print them to the screen using syntax highlighting, etc.
*/
static int insert_str(wchar_t *str)
{
	while( (*str)!=0 )
		if(!insert_char( *str++ ))
			return 0;
	return 1;
}

/**
   Calculate the length of the common prefix substring of two strings.
*/
static int comp_len( wchar_t *a, wchar_t *b )
{
	int i;
	for( i=0; a[i] != '\0' && b[i] != '\0' && a[i]==b[i]; i++)
		;
	return i;
}

/**
   Calculates information on the parameter at the specified index.
  
   \param cmd The sommand to ba analysed
   \param pos An index in the string which is inside the parameter
   \param quote An pointer to a wchar_t, which is not 0, will be used to store the type of quote this parameter has, can be either ', " or \\0.
   \param offset If not 0, get_param will store a pointer to the beginning of the parameter.
   \param string If not o, get_parm will store a copy of the parameter string as returned by the tokenizer.
   \param type If not 0, get_param will store the token type as returned by tok_last.
*/
static void get_param( wchar_t *cmd, 
					   int pos,
					   wchar_t *quote,
					   wchar_t **offset, 
					   wchar_t **string, 
					   int *type )
{
	int prev_pos=0;
	wchar_t last_quote = '\0';
	int unfinished;
	
	tokenizer tok;
	tok_init( &tok, cmd, TOK_ACCEPT_UNFINISHED );
	
	for( ; tok_has_next( &tok ); tok_next( &tok ) )
	{
		if( tok_get_pos( &tok ) > pos )
			break;
		if( tok_last_type( &tok ) == TOK_QSTRING )
		{
			last_quote = tok_last_quote( &tok );
		}
		else
			last_quote = '\0';
		if( type != 0 )
			*type = tok_last_type( &tok );
		if( string != 0 )
			wcscpy( *string, tok_last( &tok ) );
		prev_pos = tok_get_pos( &tok );
	}
	

	tok_destroy( &tok );

	
	
	wchar_t c = cmd[pos];
	cmd[pos]=0;
	int cmdlen = wcslen( cmd );
	unfinished = (cmdlen==0);	
	if( !unfinished )
	{
		unfinished = (quote != 0);
		
		if( !unfinished )
		{
			if( wcschr( L" \t\n\r", cmd[cmdlen-1] ) != 0 )
			{
				if( ( cmdlen == 1) || (cmd[cmdlen-2] != L'\\') )
				{
					unfinished=1;
				}
			}
		}
	}

	if( quote != 0 )
		*quote = last_quote;
	
	if( offset != 0 )
	{
		if( !unfinished )
		{
			while( (cmd[prev_pos] != 0) && (wcschr( L";|",cmd[prev_pos])!= 0) )
				prev_pos++;
			
			*offset = cmd + prev_pos;
		}
		else
		{
			*offset = cmd + pos;
		}
	}
	cmd[pos]=c;
}

/**
   Insert the string at the current cursor position. The function
   checks if the string is quoted or not and correctly escapes the
   string.

   \param val the string to insert
   \param is_complete Whether this completion is the whole string or
   just the common prefix of several completions. If the former, end by
   printing a space (and an end quote if the parameter is quoted).
*/
static void completion_insert( wchar_t *val, int is_complete )
{
	wchar_t *replaced;

	wchar_t quote;
	int written = 0;
	
	get_param( buff, buff_pos, &quote, 0, 0, 0 );

	if( quote == L'\0' )
	{
		replaced = expand_escape( val, 1 );
	}
	else
	{
		wchar_t *pin, *pout;
		replaced = pout = malloc( sizeof(wchar_t)*(wcslen(val) + 1) );
		for( pin=val; *pin; pin++ )
		{
			switch( *pin )
			{
				case L'\n':
				case L'\t':
				case L'\b':
				case L'\r':
					break;				
				default:
					*pout++ = *pin;
					break;
			}
		}
		*pout = 0;
	}

	if( insert_str( replaced ) )
	{
		
		if( is_complete ) /* Print trailing space since this is the only completion */
		{
			
			if( (quote) &&
				(buff[buff_pos] != quote ) ) /* This is a quoted parameter, first print a quote */
			{
				insert_char( quote );
			}
			insert_char( L' ' );
		}	
	}
	
	if( quote == L'\0' )
		free(replaced);	
}


/**
   Print the list of completions
  
   \param l The list of completions to print
   \param cols number of columns to print in
   \param width An array specifying the width of each column
   \param row_start The first row to print
   \param row_stop the row after the last row to print
   \param prefix The string to print before each completion
   \param is_quoted Whether to print the completions are in a quoted environment
*/

static void completion_print( int cols,
							  int *width, 
							  int row_start,
							  int row_stop,
							  wchar_t *prefix,
							  int is_quoted,
							  array_list *l)
{
	
	int rows = (al_get_count( l )-1)/cols+1;
	int i, j;
	int prefix_len= wcslen(prefix);

	for( i = row_start; i<row_stop; i++ )
	{
		for( j = 0; j < cols; j++ )
		{
			wchar_t *el, *el_end;
			
			if( al_get_count( l ) <= j*rows + i )
				continue;
			
			el = (wchar_t *)al_get( l, j*rows + i );
			el_end= wcschr( el, COMPLETE_SEP );
			
			set_color( highlight_get_color(HIGHLIGHT_PAGER_PREFIX),FISH_COLOR_NORMAL );
			
			writestr( prefix );

			set_color( highlight_get_color(HIGHLIGHT_PAGER_COMPLETION),FISH_COLOR_IGNORE );
			
			if( el_end == 0 )
			{
				/* We do not have a description for this completion */
				int written = 0;
				int max_written = width[j] - prefix_len - 2;
				
				if( is_quoted )
				{
					for( i=0; i<max_written; i++ )
					{
						if( !el[i] )
							break;
						writech( el[i] );
						written++;
					}					
				}				
				else
				{
					written = write_escaped_str( el, max_written );
				}

				set_color( highlight_get_color(HIGHLIGHT_PAGER_DESCRIPTION), FISH_COLOR_IGNORE );
				
				writespace( width[j]-
							written-
							prefix_len );
			}
			else
			{
				/* This completion has a description which we must also print */
				int written = 0;

				int max_written = width[j] - wcslen(el_end+1) - 4 - prefix_len - 2 + (j==cols-1?2:0);
				int i;

				/* First we must print the completion. Temporarily remove description*/
				*el_end = L'\0';
				if( is_quoted )
				{
					for( i=0; i<max_written; i++ )
					{
						if( !el[i] )
							break;
						writech( el[i] );
						written++;
					}					
//					writestr( el );
//					written = wcslen( el );
				}				
				else
				{
					written = write_escaped_str( el, max_written );
				}
				
				/* Put the desription back */
				*el_end = COMPLETE_SEP;

				/* And print it */
				set_color( highlight_get_color(HIGHLIGHT_PAGER_DESCRIPTION), FISH_COLOR_IGNORE );
				writespace( maxi( 2, 
								  width[j] 
								  - written
								  - wcslen( el_end+1 )
								  - 4 
								  - prefix_len
								  + (j==cols-1?2:0) ) );
				/* Print description */
				writestr(L"(");
				writestr( el_end+1);
				writestr(L")");
					
				if( j != cols-1)
					writestr( L"  " );
				
			}
		}
		writech( L'\n' );
	}
}

/**
   Calculates how long the specified string would be when printed on the command line. 
   
   \param str The string to be printed.
   \param is_quoted Wheter the string would be printed quoted or unquoted
*/
static int printed_length( wchar_t *str, int is_quoted )
{
	if( is_quoted )
	{
		return wcslen( str ) + ( (wcschr(str,COMPLETE_SEP)==0)?0:2 );
	}
	else
	{
		int len = 0;
		int has_description = 0;
		while( *str != 0 )
		{
			switch( *str++ )
			{
				case L'\n':
				case L'\b':
				case L'\r':
				case L'\e':
				case L'\\':
				case L'$':
				case L' ':
				case L'#':
				case L'^':
				case L'<':
				case L'>':
				case L'@':
				case L'\t':
				case L'(':
				case L')':
				case L'{':
				case L'}':
				case L'*':
				case L'?':
					len += 2;
					break;

				case COMPLETE_SEP:
					has_description = 1;
					len++;
					break;
				default:
					len++;
					break;
			}
		}
		return len + (has_description?3:0);		
	}
}


/**
   Try to print the list of completions l with the prefix prefix using
   cols as the number of columns. Return 1 if the completion list was
   printed, 0 if the terminal is to narrow for the specified number of
   columns. Always succeeds if cols is 1.

   If all the elements do not fit on the screen at once, make the list
   scrollable using the up, down and space keys to move. The list will
   exit when any other key is pressed.
*/

static int completion_try_print( int cols,
								 wchar_t *prefix,
								 int is_quoted,
								 array_list *l )
{
	int width[32];
	int i, j;
	
	int rows = (al_get_count( l )-1)/cols+1;
	int tot_width = 0;
	int prefix_len= wcslen(prefix);

	memset( width, 0, sizeof(width) );

	/* Calculated how wide the list would be */
	for( j = 0; j < cols; j++ )
	{
		for( i = 0; i<rows; i++ )
		{
			int el_w;
			wchar_t *el;
			if( al_get_count( l ) <= j*rows + i )
				continue;
				
			el = (wchar_t *)al_get( l, j*rows + i );
			el_w = printed_length( el, is_quoted );

			el_w += prefix_len;
			if( j != cols-1 )
				el_w += 2;
			width[j] = maxi( width[j], 
							 el_w );
		}
		tot_width += width[j];
	}
	if( (cols == 1) && (tot_width > termsize.ws_col) )
	{
		tot_width=termsize.ws_col;
		width[0] = termsize.ws_col;
	}

	if( tot_width <= termsize.ws_col )
	{
		/* Terminal is wide enough. Print the list! */	

		if( rows < termsize.ws_row )
		{
			/* List fits on screen. Print it and leave */
			completion_print( cols, width, 0, rows, prefix, is_quoted, l);			
		}
		else
		{
			int npos, pos = 0;
			int do_loop = 1;
			completion_print( cols, 
							  width, 
							  0,
							  termsize.ws_row-1, 
							  prefix,
							  is_quoted,
							  l);			
			/* List does not fit on screen. Print one screenfull and leave a scrollable interface */
			while(do_loop)
			{
				wchar_t msg[10];
				int percent = 100*pos/(rows-termsize.ws_row+1);
				set_color( FISH_COLOR_BLACK, 
						   highlight_get_color(HIGHLIGHT_PAGER_PROGRESS) );
				swprintf( msg, 12,
						  L" %ls(%d%%) \r", 
						  percent==100?L"":(percent >=10?L" ": L"  "), 
						  percent );
				writestr(msg);
				set_color( FISH_COLOR_NORMAL, FISH_COLOR_NORMAL );
				int c = input_readch();
				
				switch( c )
				{
					case R_HISTORY_SEARCH_BACKWARD:
						if( pos > 0 )
						{
							pos--;
							writembs(tparm( cursor_address, 0, 0));
							writembs(scroll_reverse);
							completion_print( cols,
											  width,
											  pos,
											  pos+1,
											  prefix, 
											  is_quoted,
											  l );
							writembs(tparm( cursor_address, termsize.ws_row-1, 0));
							writembs(clr_eol );
											
						}
						
						break;
						
					case R_HISTORY_SEARCH_FORWARD:
						if( pos <= (rows - termsize.ws_row ) )
						{
							pos++;
							completion_print( cols,
											  width,
											  pos+termsize.ws_row-2,
											  pos+termsize.ws_row-1,
											  prefix, 
											  is_quoted,
											  l );
						}
						break;

					case R_END_OF_HISTORY:
					case ' ':
						
						npos = mini( rows - termsize.ws_row+1,
									 pos + termsize.ws_row-1 );
						if( npos != pos )
						{
							pos = npos;
							completion_print( cols,
											  width,
											  pos,
											  pos+termsize.ws_row-1,
											  prefix, 
											  is_quoted,
											  l );
						}
						else
						{
							writembs( flash_screen );
						}
						
						break;

					case R_BEGINNING_OF_HISTORY:
						npos = maxi( 0, 
									 pos - termsize.ws_row+1 );
						
						if( npos != pos )
						{
							pos = npos;
							completion_print( cols,
											  width,
											  pos,
											  pos+termsize.ws_row-1,
											  prefix, 
											  is_quoted,
											  l );
						}
						else
						{
							writembs( flash_screen );
						}						
						break;

					default:
						input_unreadch( c );
						do_loop = 0;
						break;
								
				}
				
			}
			writembs(clr_eol);				
		}
		return 1;

	}
	return 0;
}

/**
   Handle the completion list comp. This means the following:

   - If the list is empty, flash the terminal.
   - If the list contains one element, write the whole element, and if
   the element does not end on a '/', '@', ':', or a '=', also write a trailing
   space.
   - If the list contains multiple elements with a common prefix, write
   the prefix.
   - If show_list is true, and the list contains multiple elements without 
   a common prefix, call completion_try_print to show a list of the 
   available completions.
*/


static int handle_completions( array_list *comp, int show_list )
{
	int i;
	static int completion_len=-1;
	
	if( show_list && completion_len == 0)
	{
		/* 
		   There is no common prefix in the completions, and show_list
		   is true, so we print the list 
		*/
		int len;
		wchar_t * prefix;
		wchar_t * prefix_start;
		get_param( buff, buff_pos, 0, &prefix_start, 0, 0 );

		len = &buff[buff_pos]-prefix_start;
		if( len < PREFIX_MAX_LEN )
		{
			prefix = malloc( sizeof(wchar_t)*(len+1) );
			wcsncpy( prefix, prefix_start, len );
			prefix[len]=L'\0';
		}
		else
		{
			prefix = wcsdupcat( L"...", prefix_start + (len - PREFIX_MAX_LEN+3));
			
		}

		
		{
			int is_quoted;
			
			wchar_t quote;
			get_param( buff, buff_pos, &quote, 0, 0, 0 );
			is_quoted = (quote != L'\0');
			
			writech(L'\n');

			/*
			  Try to print a list of completions. First try with five
			  columns, then four, etc. completion_try_print always
			  suceeds with one column.
			*/ 
			for( i = 5; i>0; i-- )
			{
				if( completion_try_print( i, prefix, is_quoted, comp ) )
					break;
			}
			
		}
		
		free( prefix );
		write_prompt(0);
		write_cmdline();
		move_cursor(buff_pos-buff_len);
		return 0;
	}
	completion_len =-1;
	if( al_get_count( comp ) == 0 )
	{
		if( flash_screen != 0 )
			writembs( flash_screen );
		return 0;
	}
	else if( al_get_count( comp ) == 1 )
	{
		wchar_t *comp_str = wcsdup((wchar_t *)al_get( comp, 0 ));
		wchar_t *woot = wcschr( comp_str, COMPLETE_SEP );
		if( woot != 0 )
			*woot = L'\0';
		completion_insert( comp_str, 
						   ( wcslen(comp_str) == 0 ) || 
						   ( wcschr( L"/=@:", comp_str[wcslen(comp_str)-1] ) == 0 ) );
		free( comp_str );
		return 1;
	}
	else 
	{
		wchar_t *base = wcsdup( (wchar_t *)al_get( comp, 0 ) );
		int len = wcslen( base );
		for( i=1; i<al_get_count( comp ); i++ )
		{
			int new_len = comp_len( base, (wchar_t *)al_get( comp, i ) );
			len = new_len < len ? new_len: len;
		}
		if( len > 0 )
		{
			base[len]=L'\0';
			wchar_t *woot = wcschr( base, COMPLETE_SEP );
			if( woot != 0 )
				*woot = L'\0';
			completion_insert(base, 0);
		}
		completion_len = len;						
		free( base );
		return len;
	}
}

/**
   Respond to a winch signal by checking the terminal size
*/
static void handle_winch( int sig )
{
	wchar_t tmp[64];
	if (ioctl(1,TIOCGWINSZ,&termsize)!=0) 
	{
		return;
	}	
	swprintf(tmp, 64, L"%d", termsize.ws_row );
	env_set( L"LINES", tmp, 0 );
	swprintf(tmp, 64, L"%d", termsize.ws_col );
	env_set( L"COLUMNS", tmp, 0 );
}

/**
   Interactive mode ^C handler. Respond to int signal by setting
   interupted-flag and stoping all loops and conditionals.
*/
static void handle_int( int sig )
{
	interupted=1;
	
	block_t *c = current_block;
	while( c )
	{
		c->skip=1;
		c=c->outer;
	}
	
}

/**
   Reset the terminal. This function is placed in the list of
   functions to call when exiting by using the atexit function. It
   checks whether it is the original parent process that is exiting
   and not a subshell, and if it is the parent, it restores the
   terminal.
*/
static void exit_func()
{
	if( getpid() == original_pid )
		tcsetattr(0, TCSANOW, &saved_modes);
}

/**
   Sets apropriate signal handlers.
*/
static void set_signal_handlers()
{
	struct sigaction act;
	sigemptyset( & act.sa_mask );
	act.sa_flags=0;		
	act.sa_handler=SIG_DFL;

	/*
	  First reset everything
	*/
	sigaction( SIGINT, &act, 0);
	sigaction( SIGQUIT, &act, 0);
	sigaction( SIGTSTP, &act, 0);
	sigaction( SIGTTIN, &act, 0);
	sigaction( SIGTTOU, &act, 0);
	sigaction( SIGCHLD, &act, 0);
	
	if( shell_is_interactive )
	{
		
		/* 
		   Interactive mode. Ignore interactive signals.  We are a
		   shell, we know whats best for the user. ;-)
		*/
	
		act.sa_handler=SIG_IGN;

		sigaction( SIGINT, &act, 0);
		sigaction( SIGQUIT, &act, 0);
		sigaction( SIGTSTP, &act, 0);
		sigaction( SIGTTIN, &act, 0);
		sigaction( SIGTTOU, &act, 0);
		
		act.sa_handler = &handle_int; 
		act.sa_flags = 0;
		if( sigaction( SIGINT, &act, 0) )
		{
			wperror( L"sigaction" );
			exit(1);
		}
		
		act.sa_sigaction = &job_handle_signal; 
		act.sa_flags = SA_SIGINFO;
		if( sigaction( SIGCHLD, &act, 0) )
		{
			wperror( L"sigaction" );
			exit(1);
		}
		
		act.sa_flags = 0;
		act.sa_handler= &handle_winch;
		if( sigaction( SIGWINCH, &act, 0 ) )
		{
			wperror( L"sigaction" );
			exit(1);
		}
	}
	else
	{

		/*
		  Non-interactive. Ignore interrupt, check exit status of
		  processes to determine result instead.
		*/
		act.sa_handler=SIG_IGN;
		
		sigaction( SIGINT, &act, 0);
		sigaction( SIGQUIT, &act, 0);
		
		act.sa_handler=SIG_DFL;
	}
}


/**
   Initialize data for interactive use
*/
static void reader_interactive_init()
{
	/* See if we are running interactively.  */
	pid_t shell_pgid;

	input_init();
	kill_init();	
	shell_pgid = getpgrp ();
	
	/* Loop until we are in the foreground.  */
	while (tcgetpgrp( 0 ) != shell_pgid)
	{
		kill (- shell_pgid, SIGTTIN);
	}
	
	/* Put ourselves in our own process group.  */
	shell_pgid = getpid ();
	if( getpgrp() != shell_pgid )
	{
		if (setpgid (shell_pgid, shell_pgid) < 0)
		{
			wperror( L"setpgid" );
			fwprintf (stderr,
					  L"Couldn't put the shell in its own process group");
			exit (1);
		}
	}
	
	
	/* Grab control of the terminal.  */
	if( tcsetpgrp (STDIN_FILENO, shell_pgid) )
	{
		wperror( L"tcsetpgrp" );
		exit(1);
	}
	
	
	al_init( &prompt_list );
	history_init();
	
	current_lineno=0;
	
	/* This will allocate buff, prefix_buff, etc... */
	check_size();

	buff[0] = prefix_buff[0] = '\0';

	handle_winch( 0 );                /* Set handler fow window change events */	
	tcgetattr(0,&shell_modes);        /* get the current terminal modes */
	saved_modes = shell_modes;          /* save a copy so we can reset them */
	
	shell_modes.c_lflag &= ~ICANON;   /* turn off canonical mode */
	shell_modes.c_lflag &= ~ECHO;     /* turn off echo mode */
    shell_modes.c_cc[VMIN]=1;
    shell_modes.c_cc[VTIME]=0;
	
    if( tcsetattr(0,TCSANOW,&shell_modes))      /* set the new modes */
    {
        wperror(L"tcsetattr");
        exit(1);
    }
	
	original_pid = getpid();
	
	if( atexit( &exit_func ) )
		fwprintf( stderr, L"Cannot set exit function\n" );
	
	env_set( L"_", L"fish", ENV_GLOBAL );
}

/**
   Destroy data for interactive use
*/
static void reader_interactive_destroy()
{
	kill_destroy();	
	al_foreach( &prompt_list, (void (*)(const void *))&free );
	al_destroy( &prompt_list );
	history_destroy();

	free(buff);
	free(color);
	free(new_color);
	free(prefix_buff);	
	
	writestr( L"\n" );
	set_color( FISH_COLOR_RESET, FISH_COLOR_RESET );
	input_destroy();
}


void reader_sanity_check()
{
	if( shell_is_interactive)
	{
		if(!( buff_pos <= buff_len ))
			sanity_lose();
		if(!( buff_pos >=0 ))
			sanity_lose();
		if(!( buff_len >=0 ))
			sanity_lose();
		if(!( buff_len == wcslen( buff ) ))
			sanity_lose();
	}
}

/**
   Set the specified string from the history as the current buffer. Do
   not modify prefix_len.
*/
static void handle_history( const wchar_t *new_str )
{
	buff_len = wcslen( new_str );
	check_size();
	wcscpy( buff, new_str );
	buff_pos=buff_len=wcslen(buff);
	repaint();					
}

/**
   Move buffer position one word or erase one word. This function updates both
   the internal buffer an the screen. It is used by M-left, M-right
   and ^W to do block movement or block erase.

   \param dir Direction to move/erase. 0 means move left, 1 means move right.
   \param erase Whether to erase the characters alopng the way or only move past them.
   
*/
static void move_word( int dir, int erase )
{
	int i;
	int had_char=0;
	int end_buff_pos=buff_pos;
	while( 1 )
	{
		if( !dir )
		{
			if( end_buff_pos == 0 )
				break;
			end_buff_pos--;
		}
		else
		{
			if( end_buff_pos == buff_len )
				break;
			end_buff_pos++;
		}
		

		if( buff[end_buff_pos] == L' ' )
		{
			if( had_char )
			{
				break;
			}
		}
		else
		{
			had_char = 1;
		}
	}
	if( erase )
	{
		int remove_count = abs(buff_pos - end_buff_pos);
		int first_char = mini( buff_pos, end_buff_pos );	
		wchar_t *woot = wcsndup( buff + first_char, remove_count);
//		fwprintf( stderr, L"Remove from %d to %d\n", first_char, first_char+remove_count );
		
		kill_add( woot );
		free( woot );
		memmove( buff + first_char, buff + first_char+remove_count, sizeof(wchar_t)*(buff_len-first_char-remove_count) );
		buff_len -= remove_count;
		buff_pos = first_char;
		buff[buff_len]=0;
		
		repaint();		
	}
	else
	{
		move_cursor(end_buff_pos-buff_pos);
		buff_pos = end_buff_pos;
		check_colors();
	}
	history_reset();
}


wchar_t *reader_get_buffer()
{
	return buff;
	
}

int reader_get_cursor_pos()
{
	return buff_pos;
}



void reader_run_command( wchar_t *cmd )
{
	set_color(FISH_COLOR_NORMAL, FISH_COLOR_NORMAL);
	while( 1 )
	{
		if(	tcsetattr(0,TCSANOW,&saved_modes) )
		{
			if( errno != EINTR )
			{
				fwprintf( stderr, L"Could not set terminal mode for new job\n" );
				wperror( L"tcsetattr" );
				break;
			}
		}
		else
			break;
	}					
					
	wchar_t *ft = tok_first( cmd );
	if( ft != 0 )
		env_set( L"_", ft, ENV_GLOBAL );
	free(ft);
				   
	reader_write_title();
					
	if( eval( cmd ) == 0 )
	{
		job_do_notification();
	}

	while( 1 )
	{
		if(	tcsetattr(0,TCSANOW,&shell_modes) )
		{
			if( errno != EINTR )
			{
				fwprintf( stderr, L"Could not set terminal mode for shell\n" );
				wperror( L"tcsetattr" );
				break;
			}
		}
		else
			break;
	}

	env_set( L"_", L"fish", ENV_GLOBAL );
	handle_winch( 0 );
	
#ifdef HAVE__PROC_SELF_STAT
	proc_update_jiffies();
#endif

	write_prompt(1);
	if( cmd != buff )
	{
		repaint();
	}
}


/**
   Read interactively. Read input from stdin while providing editing facilities.
*/
static int read_i()
{
	wchar_t c;	
	int i;
	int last_char=0, yank=0;
	wchar_t *yank_str;
	array_list comp;
	int tmp;
	int comp_empty=1;		

	check_size();
	prefix_buff[0]=buff[0]='\0';
	
	reader_interactive_init();
	
	al_init( &comp );
	
	write_prompt(1);                  
	while( (!end_loop) && (!sanity_check()) )
	{

		/*
		  Save the terminal status so we know if we have to redraw
		*/
		
		/* 
		   Sometimes strange input sequences seem to generate a zero
		   byte. I belive these simply mean a character was pressed
		   but it should be ignored. (Example: Trying to add a tilde
		   (~) to digit)
		*/

		reader_save_status();
		while( (c=input_readch()) == 0 )
			;
		reader_check_status();

		if( (last_char == R_COMPLETE) && (c != R_COMPLETE) && (!comp_empty) )
		{
			al_foreach( &comp, (void (*)(const void *))&free );
			al_truncate( &comp, 0 );
			comp_empty = 1;
		}

		if( last_char != R_YANK && last_char != R_YANK_POP )
			yank=0;
		switch (c)
		{
			/* go to beginning of line*/

			case R_BEGINING_OF_LINE:
				for( i=0; i<buff_pos; i++ )
					writembs( cursor_left );
				buff_pos = 0;				
				break;
					
				/* go to EOL*/
			case R_END_OF_LINE:
				for( i=buff_pos; i<buff_len; i++ )
					writembs( cursor_right );
				buff_pos = buff_len;				
				break;

			case R_EXPLAIN:
			{
				int i;
					
				array_list l;
				al_init( &l );
				highlight( buff, new_color, buff_pos, &l );
				if( al_get_count( &l ) > 0 )
				{
					set_color(FISH_COLOR_NORMAL, FISH_COLOR_NORMAL);
					writestr( L"\n" );
					for( i=0; i<al_get_count( &l ); i++ )
					{
						wchar_t *str = (wchar_t *)al_get( &l, i );
						writestr( str );
						writestr( L"\n" );
					}

					repaint();
					al_foreach( &l, (void (*)(const void *))&free );
				}
					
				al_destroy( &l );
					
				break;
			}
				
			/* complete */
			case R_COMPLETE:
//					fwprintf( stderr, L"aaa\n" );
				if( comp_empty )
				{
						
					wchar_t *buffcpy = wcsdup( buff );
					wchar_t *begin, *end;

					while( 1 )
					{
						int bc, ec;
						
						if( expand_locate_subshell( buffcpy, 
													&begin, 
													&end,
													1 ) <= 0)
						{
							begin=buffcpy;
							end = buffcpy + wcslen(buff);
							break;
						}
						bc = begin-buffcpy;
						ec = end-buffcpy;
						if(( bc < buff_pos ) && (ec >= buff_pos) )
						{
							end--;
							begin++;
								
							//fwprintf( stderr, L"Subshell!\n" );
							break;
						}
						*begin=0;
					}
						
					int len = buff_pos-(begin-buffcpy); 
					wcsncpy( buffcpy, 
							 buff+(begin-buffcpy), 
							 len );
					buffcpy[len]=0;
						
					//fwprintf( stderr, L"String is %ls\n", buffcpy );
												
					reader_save_status();
					complete( buffcpy, &comp );					
					reader_check_status();
						
					sort_list( &comp );
					remove_duplicates( &comp );
						
					free( buffcpy );
				}
				if( (comp_empty =
					 handle_completions( &comp, 
										 last_char == R_COMPLETE ) ) )
				{
					al_foreach( &comp, (void (*)(const void *))&free );
					al_truncate( &comp, 0 );
				}

				break;

				/* kill*/
			case R_KILL_LINE:
			{
				kill_add( &buff[buff_pos] );
				buff_len = buff_pos;
				buff[buff_len]=L'\0';
				
				writembs(clr_eol);				
				wcscpy(prefix_buff,buff);
				break;
			}
			
			case R_BACKWARD_KILL_LINE:
			{
				wchar_t prev = buff[buff_pos];
				buff[buff_pos]=0;
				kill_add( buff );
				buff[buff_pos]=prev;
				buff_len = wcslen(buff +buff_pos);
				memmove( buff, buff +buff_pos, sizeof(wchar_t)*buff_len );
				buff[buff_len]=L'\0';
				buff_pos=0;
				repaint();
				wcscpy(prefix_buff,buff);
				break;
			}
			
			case R_KILL_WHOLE_LINE:
			{
				kill_add( buff );
				buff_len = buff_pos = 0;
				buff[buff_len]=L'\0';
				repaint();
				wcscpy(prefix_buff,buff);
				break;
			}
			
			/* yank*/
			case R_YANK:
				yank_str = kill_yank();
				insert_str( yank_str );
				yank = wcslen( yank_str );
				wcscpy(prefix_buff,buff);
				break;
				
				/* rotate killring*/	
			case R_YANK_POP:
				if( yank )
				{
					for( i=0; i<yank; i++ )
						remove_backward();
						
					yank_str = kill_yank_rotate();
					insert_str(yank_str);
					yank = wcslen(yank_str);
				}					
				break;

				/* Escape, do nothing*/
			case L'\e':
				break;

				/* delete backward*/
			case R_BACKWARD_DELETE_CHAR:	
				remove_backward();
				break;
					
				/* delete forward*/
			case R_DELETE_CHAR:
				remove_forward();
				break;

				/* ^D, exit, but only if nothing has been written*/
			case 4:
				if( buff_len == 0 )
				{
					end_loop=1;
				}
				break;						

				/* Newline, evaluate*/
			case L'\n':

				buff[buff_len]=L'\0';
				writestr( L"\n" );
				if( parser_test( buff, 1 ) )
				{
					repaint();
					break;
				}
					
				current_lineno++;
					
				wcscpy(prefix_buff,L"");
				history_reset();
				history_add( buff );
				if( buff_len == 0 )
				{
					repaint();
					break;
				}
					
				reader_run_command( buff );
					
				buff_pos=buff_len=0;		
				buff[buff_len]=L'\0';

				break;
					
				/* History up*/
			case R_HISTORY_SEARCH_BACKWARD:
//					fwprintf( stderr, L"Search history for \'%ls\' %d long\n", prefix_buff, wcslen(prefix_buff) );

				handle_history(history_prev_match(prefix_buff));
				break;
				
				/* History down*/
			case R_HISTORY_SEARCH_FORWARD:				
				handle_history(history_next_match(prefix_buff));
				break;
					
				/* Move left*/
			case R_BACKWARD_CHAR:
				if( buff_pos > 0 )
				{
					buff_pos--;
					writembs( cursor_left );
					check_colors();
				}				
				break;

				/* Move right*/
			case R_FORWARD_CHAR:
				if( buff_pos < buff_len )
				{
					buff_pos++;
					writembs( cursor_right );
					check_colors();
					
				}
				break;
					
			case R_DELETE_LINE:
				buff[0]=0;
				buff_len=0;
				buff_pos=0;
				repaint();
				
				/* kill one word left */
			case R_BACKWARD_KILL_WORD:
				move_word(0,1);
				break;
					
				/* kill one word right */
			case R_KILL_WORD:
				move_word(1,1);
				break;
					
				/* move one word left*/
			case R_BACKWARD_WORD:
				move_word(0,0);
				break;
					
				/* move one word right*/
			case R_FORWARD_WORD:
				move_word( 1,0);
				break;
					
				/* Other, if a normal character, we add it to the command */
			default:
				if( (c< WCHAR_END) && (c>26) )
				{
					history_reset();
					insert_char( c );
				}
				break;
		}
		last_char = c;
		if( end_loop)
		{
			if( first_job != 0 )
			{
				writestr(L"There are stopped jobs\n");
				write_prompt(0);				
				end_loop = 0;
			}
		}	
	}
	
	al_destroy( &comp );
	reader_interactive_destroy();
	return 0;
}


/**
   Read non-interactively.  Read input from stdin without displaying
   the prompt, using syntax highlighting. This is used for reading
   scripts and init files.
*/
static int read_ni()
{
	wchar_t *buff=0;
	wchar_t *acc_buff=0;
	int buff_len=0;
	FILE *in_stream;
	int des = dup( 0 );
	current_lineno=0;
	if (des == -1)
	{
		wperror( L"dup" );
		return 1;
	}
	
	in_stream = fdopen( des, "r" );
	if( in_stream != 0 )
	{
		while( (!feof( in_stream )) && (!sanity_check()) && (!end_loop))
		{
			switch( fgetws2( &buff, &buff_len, in_stream ) )
			{
				case -1:
					wperror( L"fgetws2 (read_ni)" );
					free( buff );
					return 1;
				default:
				{
					int test_val;
				
					current_lineno++;
					if( acc_buff == 0 )
						acc_buff = buff;
					else
					{ 
						wchar_t *tmp = wcsdupcat2( acc_buff, L"\n", buff, 0 );
						if( acc_buff != buff )
							free( acc_buff );
						acc_buff = tmp;
					}
					
					test_val = parser_test( acc_buff, 0 );
					
					if( !test_val ) 
					{
						eval( acc_buff );
						if( acc_buff != buff )
							free( acc_buff );
						acc_buff = 0;
					}
					else 
					{
						if( test_val & 1 )
						{
							test_val = parser_test( acc_buff, 1 );
							if( acc_buff != buff )
								free( acc_buff );
							acc_buff = 0;
						}
						else
						{
							if( acc_buff == buff )
							{
								acc_buff = wcsdup( acc_buff );
							}
						}
					}
				}				
			}
		}
		if( acc_buff )
		{
			eval( acc_buff );
			if( acc_buff != buff )
				free( acc_buff );
		}
		free(buff);
		
		if(	fclose( in_stream ))
		{
			wperror( L"fclose" );
			free( buff );
			return 1;
		}

		return 0;
	}
	else
	{
		wperror( L"fdopen" );
		free( buff );
		return 1;
	}	
}


int reader_read()
{
	int res;	
	/* 
	   If reader_read is called recursively through the '.' builtin,
	   we need to preserve shell_is_interactive, so we save the
	   original state. We also update the signal handlers.
	*/
	int shell_was_interactive = shell_is_interactive;
	shell_is_interactive = isatty(STDIN_FILENO);
	set_signal_handlers();
	
	if( shell_is_interactive )
		res = read_i();
	else 
		res = read_ni();
	
	/* 
	   If the exit command was called in a script, only exit the
	   script, not the program
	*/
	end_loop = 0;	

	shell_is_interactive = shell_was_interactive;
	set_signal_handlers();
	return res;
}
